# -*- coding: utf-8 -*-
"""
.. module:: cindex
    :synopsis: Calculate the cindex.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict


def compute_cindex(df, colgroupby, colcountby, show_progress=False):
    """
    Calculate the cindex for each group in the DataFrame (the number of citations to the maximum cited publication).
    See :cite:`Waltman2008index` for detailed definition.

    Parameters
    ----------
    :param df : DataFrame
        A DataFrame with the citation information for each Author.

    :param colgroupby : str
        The DataFrame column with Author Ids.

    :param colcountby : str
        The DataFrame column with Citation counts for each publication.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: colgroupby, 'Cindex'

        """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='cindex', disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colgroupby)+'Cindex']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].max(qfactor).to_frame().reset_index().rename(columns=newname_dict)

def compute_pindex(df, colgroupby, colcountby, show_progress=False):
    """
    Calculate the pindex for each group in the DataFrame (the number of publications with >0 citations).  
    See :cite:`Waltman2008index` for detailed definition.

    Parameters
    ----------
    :param df : DataFrame
        A DataFrame with the citation information for each Author.

    :param colgroupby : str
        The DataFrame column with Author Ids.

    :param colcountby : str
        The DataFrame column with Citation counts for each publication.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: colgroupby, 'Pindex'

        """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='pindex', disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colgroupby)+'Pindex']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].astype(bool).sum(axis=0).to_frame().reset_index().rename(columns=newname_dict)




# -*- coding: utf-8 -*-
"""
.. module:: all
    :synopsis: easy interface to all of pyscisci

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

from pyscisci.utils import *
from pyscisci.methods.publication import *
from pyscisci.methods.journal import *
from pyscisci.methods.author import *
from pyscisci.methods.referencestrength import *
from pyscisci.datasource.readwrite import load_preprocessed_data, append_to_preprocessed
from pyscisci.network import *
from pyscisci.sparsenetworkutils import *
from pyscisci.nlp import *
from pyscisci.datasource.MAG import MAG
from pyscisci.datasource.WOS import WOS
from pyscisci.datasource.DBLP import DBLP
from pyscisci.datasource.APS import APS
from pyscisci.datasource.PubMed import PubMed
from pyscisci.datasource.OpenAlex import OpenAlex
from pyscisci.datasource.CustomDB import CustomDB
from pyscisci.filter import *
from pyscisci.visualization import *



import os
import sys
import json
import gzip
import glob

import pandas as pd
import numpy as np
from nameparser import HumanName
import requests
import ftplib
from lxml import etree
from io import BytesIO

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


from pyscisci.datasource.readwrite import load_preprocessed_data, load_int, load_float, load_html_str, load_xml_text
from pyscisci.database import BibDataBase

class PubMed(BibDataBase):
    """
    Base class for PubMed Medline interface.

    Notes
    -------
    ~ PubMed comes as >1000 compressed XML files.
    ~ The PMID is renamed PublicationId to be consistent with the rest of pySciSci.
    ~ PubMed does not disambiguate Authors.
    ~
    """


    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter=None, 
        enable_dask=False, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, enable_dask, show_progress)

        self.PublicationIdType = int
        self.AffiliationIdType = int
        self.AuthorIdType = str
        self.JournalIdType = str


    def _blank_pubmed_publication(self, PublicationId = 0):
        record = {}
        record['PublicationId'] = PublicationId
        record['Title'] = ''
        record['Year'] = 0
        record['Volume'] = 0
        record['Issue'] = ''
        record['Pages'] = ''
        record['JournalId'] = ''
        record['TeamSize'] = 0
        record['Month'] = 1
        record['Day'] = 1
        record['ISSN'] = ''
        record['Doi'] = ''
        record['PMCID'] = ''

        return record

    def _blank_pubmed_author(self):
        record = {}
        record['PublicationId'] = ''
        record['FullName'] = ''
        record['FirstName'] = ''
        record['LastName'] = ''
        record['Affiliations'] = ''
        record['ORCID'] = ''
        record['AuthorSequence'] = 0

        return record

    def _blank_pubmed_grant(self):
        record = {}
        record['PublicationId'] = ''
        record['GrantID'] = ''
        record['Acronym'] = ''
        record['Agency'] = ''
        record['Country'] = ''

        return record

    def _save_dataframes(self, ifile, publication, paa, pub2ref, pub2field, pub2abstract, pub2grant):

        publication = pd.DataFrame(publication)
        publication['PublicationId'] = publication['PublicationId'].astype(int)
        publication['Year'] = publication['Year'].astype(int)
        publication['Month'] = publication['Month'].astype(int)
        publication['Day'] = publication['Day'].astype(int)
        publication['Volume'] = pd.to_numeric(publication['Volume'])
        publication['TeamSize'] = publication['TeamSize'].astype(int)
        fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
        self.save_data_file(publication, fname, key =self.path2pub)


        paa = pd.DataFrame(paa)
        paa['AuthorSequence'] = paa['AuthorSequence'].astype(int)
        fname = os.path.join(self.path2database, self.path2paa, '{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
        self.save_data_file(paa, fname, key =self.path2paa)

        pub2field = pd.DataFrame(pub2field, columns = ['PublicationId', 'FieldId'])
        pub2field['PublicationId'] = pub2field['PublicationId'].astype(int)
        fname = os.path.join(self.path2database, self.path2pub2field, '{}{}.{}'.format(self.path2pub2field, ifile, self.database_extension))
        self.save_data_file(pub2field, fname, key =self.path2pub2field)

        pub2ref = pd.DataFrame(pub2ref, columns = ['CitedPublicationId', 'CitingPublicationId', 'Citation'])
        fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
        self.save_data_file(pub2ref, fname, key =self.path2pub2ref)

        pub2abstract = pd.DataFrame(pub2abstract, columns = ['PublicationId', 'Abstract'])
        pub2abstract['PublicationId'] = pub2abstract['PublicationId'].astype(int)
        fname = os.path.join(self.path2database, self.path2pub2abstract, '{}{}.{}'.format(self.path2pub2abstract, ifile, self.database_extension))
        self.save_data_file(pub2abstract, fname, key =self.path2pub2abstract)

        pub2grant = pd.DataFrame(pub2grant)
        #pub2grant['PublicationId'] = pub2grant['PublicationId'].astype(int)
        fname = os.path.join(self.path2database, 'pub2grant', '{}{}.{}'.format('pub2grant', ifile, self.database_extension))
        self.save_data_file(pub2grant, fname, key ='pub2grant')

    def preprocess(self, xml_directory = 'RawXML', process_name=True, num_file_lines=10**6, show_progress=True,rewrite_existing = False):
        """
        Bulk preprocess of the PubMed raw data.

        Parameters
        ----------
        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.

        num_file_lines: int, default 10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.

        show_progress: bool, default True
            Show progress with processing of the data.

        rewrite_existing: bool, default False
            If True, rewrites the files in the data directory
        """

        if show_progress:
            print("Starting to preprocess the PubMed database.")

        for hier_dir_type in [self.path2pub, self.path2paa, self.path2pub2field, self.path2pub2ref, self.path2fieldinfo, self.path2pub2abstract, 'pub2grant']:

            if not os.path.exists(os.path.join(self.path2database, hier_dir_type)):
                os.mkdir(os.path.join(self.path2database, hier_dir_type))


        xmlfiles = sorted([fname for fname in os.listdir(os.path.join(self.path2database, xml_directory)) if '.xml' in fname])

        # read dtd - this takes
        path2database = self.path2database # remove self to use inside of this class
        class DTDResolver(etree.Resolver):
            def resolve(self, system_url, public_id, context):
                return self.resolve_filename(os.path.join(path2database, system_url), context)
        parser = etree.XMLParser(load_dtd=True, resolve_entities=True)

        pub2year = {}
        fieldinfo = {}

        #xmlfiles = ['pubmed22n1114.xml.gz']
        
        ifile = 0
        for xml_file_name in tqdm(xmlfiles, desc='PubMed xml files', leave=True, disable=not show_progress):

            # check if the xml file was already parsed
            dest_file_name = os.path.join(self.path2database, self.path2pub,'publication{}.hdf'.format(ifile))
            if not rewrite_existing and os.path.isfile(dest_file_name):
                ifile+=1
                continue

            publication = []
            paa = []
            pub2field = []
            pub2ref = []
            pub2abstract = []
            pub2grant = []

            xmltree = etree.parse(os.path.join(self.path2database, xml_directory, xml_file_name), parser)

            all_pubmed_articles = xmltree.findall("/PubmedArticle")
            
            for article_bucket in all_pubmed_articles:

                medline = article_bucket.find("MedlineCitation")

                # scrape the publication information
                PublicationId = load_int(load_xml_text(medline.find('PMID')))
                pub_record = self._blank_pubmed_publication(PublicationId)

                article = medline.find("Article")
                pub_record['Title'] = load_html_str(load_xml_text(article.find('ArticleTitle')))
                if article.find('Pagination') == None:
                    pub_record['Pages'] = None
                else:
                    pub_record['Pages'] = load_html_str(load_xml_text(article.find('Pagination').find("MedlinePgn")))

                journal = article.find("Journal")
                pub_record['JournalId'] = load_html_str(load_xml_text(journal.find("Title")))
                pub_record['Volume'] = load_int(load_xml_text(journal.find("JournalIssue").find("Volume")))
                pub_record['Issue'] = load_int(load_xml_text(journal.find("JournalIssue").find("Issue")))
                pub_record['ISSN'] = load_html_str(load_xml_text(journal.find("ISSN")))

                history = article_bucket.find("PubmedData/History")
                if not history is None:
                    pdate = history.find('PubMedPubDate')
                    if not pdate is None:
                        pub_record['Year'] = load_int(load_xml_text(pdate.find("Year")))
                        pub_record['Month'] = load_int(load_xml_text(pdate.find("Month")))
                        pub_record['Day'] = load_int(load_xml_text(pdate.find("Day")))


                if pub_record['Year'] > 0:
                    pub2year[PublicationId] = pub_record['Year']

                article_ids = article_bucket.find("PubmedData/ArticleIdList")
                if article_ids is not None:
                    doi = article_ids.find('ArticleId[@IdType="doi"]')
                    pub_record['Doi'] = load_xml_text(doi)

                    pmcid = article_ids.find('ArticleId[@IdType="pmc"]')
                    pub_record['PMCID'] = load_xml_text(pmcid)


                pub2abstract.append({'PublicationId':PublicationId, 'Abstract':load_xml_text(article.find("Abstract"))})

                grant_list = article.find('GrantList')
                if not grant_list is None:
                    for grant in grant_list:
                        grant_record = self._blank_pubmed_grant()
                        grant_record['PublicationId'] = PublicationId

                        grant_record['GrantID'] = load_xml_text(grant.find("GrantID"))
                        grant_record['Acronym'] = load_xml_text(grant.find("Acronym"))
                        grant_record['Agency'] = load_xml_text(grant.find("Agency"))
                        grant_record['Country'] = load_xml_text(grant.find("Country"))

                        pub2grant.append(grant_record)


                author_list = article.find('AuthorList')

                if not author_list is None:
                    for seq, author in enumerate(author_list.findall('Author')):
                        author_record = self._blank_pubmed_author()

                        author_record['PublicationId'] = PublicationId
                        author_record['FirstName'] = load_html_str(load_xml_text(author.find("ForeName")))
                        author_record['LastName'] = load_html_str(load_xml_text(author.find("LastName")))
                        author_record['FullName'] = author_record['FirstName'] + ' ' + author_record['LastName']
                        author_record['ORCID'] = load_xml_text(author.find('Identifier[@Source="ORCID"]'))
                        
                        if author.find("AffiliationInfo/Affiliation") is not None:
                            author_record['Affiliations'] = load_html_str(load_xml_text(author.find("AffiliationInfo/Affiliation")))
                            author_record['Affiliations'] = author_record['Affiliations'].replace("For a full list of the authors' affiliations please see the Acknowledgements section.","")

                        author_record['AuthorSequence'] = seq+1

                        paa.append(author_record)

                    pub_record['TeamSize'] = seq + 1

                meshterms = medline.find("MeshHeadingList")

                if meshterms is not None:
                    for term in meshterms.getchildren():
                        ui = term.find("DescriptorName").attrib.get("UI", "")
                        if len(ui)>0:
                            pub2field.append([PublicationId, ui])
                            fieldinfo[ui] = [load_xml_text(term.find("DescriptorName")), 'mesh']

                chemicals = medline.find("ChemicalList")
                if chemicals is not None:
                    for chemical in chemicals.findall("Chemical"):
                        ui = chemical.find("NameOfSubstance").attrib.get("UI", "")
                        if len(ui)>0:
                            pub2field.append([PublicationId, ui])
                            fieldinfo[ui] = [load_xml_text(chemical.find("NameOfSubstance")), 'chem']

                references = article_bucket.find("PubmedData/ReferenceList")
                if not references is None:
                    for ref in references.findall("Reference"):
                        citation = load_xml_text(ref.find("Citation"))
                        if not ref.find('ArticleIdList') is None:
                            pmid = load_int(load_xml_text(ref.find('ArticleIdList').find('ArticleId[@IdType="pubmed"]')))
                        else:
                            pmid = ""
                        pub2ref.append([PublicationId, pmid, citation])

                publication.append(pub_record)

            self._save_dataframes(ifile, publication, paa, pub2ref, pub2field, pub2abstract, pub2grant)
            ifile += 1

        # if rewriting
        #dest_file_name = os.path.join(self.path2database, self.path2fieldinfo,'fieldinfo0.hdf')
        if rewrite_existing:
            # save field info dictionary
            mesh_id_list = list(fieldinfo.values())
            for i, j in enumerate(fieldinfo.keys()):
                mesh_id_list[i].insert(0, j)

            fieldinfo = pd.DataFrame(mesh_id_list, columns = ['FieldId', 'FieldName', 'FieldType'], dtype=int)
            fname = os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format(self.path2fieldinfo, 0, self.database_extension))
            self.save_data_file(fieldinfo, fname, key =self.path2fieldinfo)

        with gzip.open(os.path.join(self.path2database, 'pub2year.json.gz'), 'w') as outfile:
            outfile.write(json.dumps(pub2year).encode('utf8'))


    def download_from_source(self, source_url='ftp.ncbi.nlm.nih.gov', dtd_url = 'https://dtd.nlm.nih.gov/ncbi/pubmed/out/pubmed_190101.dtd',
        rewrite_existing = False, show_progress=True):
        """
        Download the Pubmed raw xml files and the dtd formating information from [PubMed](https://www.nlm.nih.gov/databases/download/pubmed_medline.html).
            1. pubmed/baseline - the directory containing the baseline compressed xml files
            2. pubmed_190101.dtd - the dtd containing xml syntax

        The files will be saved to the path specified by `path2database` into RawXML.

        Parameters
        ----------
        source_url: str, default 'ftp.ncbi.nlm.nih.gov'
            The base url for the ftp server from which to download.

        dtd_url: str, default 'pubmed_190101.dtd'
            The url for the dtd file.

        rewrite_existing: bool, default False
            If True, overwrite existing files or if False, only download any missing files.

        show_progress: bool, default True
            Show progress with processing of the data.

        """

        FTP_USER = "anonymous"
        FTP_PASS = ""

        ftp = ftplib.FTP(source_url, FTP_USER, FTP_PASS)
        ftp.encoding = "utf-8"
        ftp.cwd("pubmed/baseline/")

        files2download = sorted([fname for fname in ftp.nlst() if '.xml.gz' in fname and not '.md5' in fname])

        if not os.path.exists(os.path.join(self.path2database, 'RawXML')):
            os.mkdir(os.path.join(self.path2database, 'RawXML'))

        if not rewrite_existing:
            files_already_downloaded = os.listdir(os.path.join(self.path2database, 'RawXML'))
            files2download = [fname for fname in files2download if not fname in files_already_downloaded]

        for xml_file_name in tqdm(files2download, disable=not show_progress):
            with open(os.path.join(self.path2database, 'RawXML', xml_file_name), "wb") as outfile:
                ftp.retrbinary('RETR %s' % xml_file_name, outfile.write)

        with open(os.path.join(self.path2database, 'RawXML', 'pubmed_190101.dtd'), 'w') as outfile:
            outfile.write(requests.get(dtd_url).content.decode('utf-8'))

        ftp.quit()


    def parse_affiliations(self, preprocess = False):
        raise NotImplementedError("PubMed artciles are stored with all information in an xml file.  Run preprocess to parse the file.")

    def parse_publications(self, xml_directory = 'RawXML',preprocess = True, num_file_lines=10**7,rewrite_existing = False):
        """
        Parse the PubMed publication raw data.
        
        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.
        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.
        num_file_lines: int, default 5*10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.
        show_progress: bool, default True
            Show progress with processing of the data.
        
        Returns
        ----------
        DataFrame
            Publication metadata DataFrame.
        """

        # process publication files through xml
        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2pub)):
                os.mkdir(os.path.join(self.path2database, self.path2pub))

            xmlfiles = sorted([fname for fname in os.listdir(os.path.join(self.path2database, xml_directory)) if '.xml' in fname])

            # read dtd - this takes
            path2database = self.path2database # remove self to use inside of this class
            class DTDResolver(etree.Resolver):
                def resolve(self, system_url, public_id, context):
                    return self.resolve_filename(os.path.join(path2database, system_url), context)

            parser = etree.XMLParser(load_dtd=True, resolve_entities=True)

            ifile = 0
            for xml_file_name in tqdm(xmlfiles, desc='PubMed publication xml files', leave=True, disable=not show_progress):

                # check if the xml file was already parsed
                dest_file_name = os.path.join(self.path2database, self.path2pub,'{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
                if not rewrite_existing and os.path.isfile(dest_file_name):
                    ifile+=1
                    continue

                publication = []

                all_pubmed_articles = xmltree.findall("/PubmedArticle")

                for article_bucket in all_pubmed_articles:

                    medline = article_bucket.find("MedlineCitation")

                    # scrape the publication information
                    PublicationId = load_int(load_xml_text(medline.find('PMID')))

                    pub_record = self._blank_pubmed_publication(PublicationId)

                    article = medline.find("Article")
                    pub_record['Title'] = load_html_str(load_xml_text(article.find('ArticleTitle')))
                    if article.find('Pagination') == None:
                        pub_record['Pages'] = None
                    else:
                        pub_record['Pages'] = load_html_str(load_xml_text(article.find('Pagination').find("MedlinePgn")))

                    journal = article.find("Journal")
                    pub_record['JournalId'] = load_html_str(load_xml_text(journal.find("Title")))
                    pub_record['Volume'] = load_int(load_xml_text(journal.find("JournalIssue").find("Volume")))
                    pub_record['Issue'] = load_int(load_xml_text(journal.find("JournalIssue").find("Issue")))
                    pub_record['ISSN'] = load_html_str(load_xml_text(journal.find("ISSN")))

                    history = article_bucket.find("PubmedData/History")
                    if not history is None:
                        pdate = history.find('PubMedPubDate')
                        if not pdate is None:
                            pub_record['Year'] = load_int(load_xml_text(pdate.find("Year")))
                            pub_record['Month'] = load_int(load_xml_text(pdate.find("Month")))
                            pub_record['Day'] = load_int(load_xml_text(pdate.find("Day")))

                    article_ids = article_bucket.find("PubmedData/ArticleIdList")
                    if article_ids is not None:
                        doi = article_ids.find('ArticleId[@IdType="doi"]')
                        pub_record['Doi'] = load_xml_text(doi)


                    author_list = article.find('AuthorList')

                    if not author_list is None:
                        pub_record['TeamSize'] = len(author_list.findall('Author'))

                    publication.append(pub_record)

                # save publication dataframe
                publication = pd.DataFrame(publication)
                publication['PublicationId'] = publication['PublicationId'].astype(int)
                publication['Year'] = publication['Year'].astype(int)
                publication['Month'] = publication['Month'].astype(int)
                publication['Day'] = publication['Day'].astype(int)
                publication['Volume'] = pd.to_numeric(publication['Volume'])
                publication['TeamSize'] = publication['TeamSize'].astype(int)
                fname = os.path.join(self.path2database, self.path2pub, '{}{}.{}'.format(self.path2pub, ifile, self.database_extension))
                self.save_data_file(publication, fname, key =self.path2pub)

                ifile += 1

        ## load publication dataframe into a large file
        pub_files_list = glob.glob(os.path.join(self.path2database, self.path2pub, + '{}*.{}'.format(self.path2pub, self.database_extension)))

        pub = pd.DataFrame()

        print("Parsing files...")
        for tmp_pub in tqdm(paa_files_list, desc='PubMed author files', leave=True, disable=not show_progress):
            pub = pub.append(self.read_data_file(tmp_pub), ignore_index = True)

        return pub

    def parse_references(self, xml_directory='RawXML',preprocess = True, num_file_lines=10**7, rewrite_existing=False,show_progress=True):
        """
        Parse the PubMed References raw data.
        
        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.
        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.
        num_file_lines: int, default 5*10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.
        show_progress: bool, default True
            Show progress with processing of the data.
        
        Returns
        ----------
        DataFrame
            Citations DataFrame.
        """

        # process author files through xml
        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.path2pub2ref)):
                os.mkdir(os.path.join(self.path2database, self.path2pub2ref))

            xmlfiles = sorted([fname for fname in os.listdir(os.path.join(self.path2database, xml_directory)) if '.xml' in fname])

            # read dtd - this takes
            path2database = self.path2database # remove self to use inside of this class
            class DTDResolver(etree.Resolver):
                def resolve(self, system_url, public_id, context):
                    return self.resolve_filename(os.path.join(path2database, system_url), context)

            parser = etree.XMLParser(load_dtd=True, resolve_entities=True)

            ifile = 0
            for xml_file_name in tqdm(xmlfiles, desc='PubMed reference xml files', leave=True, disable=not show_progress):

                xmltree = etree.parse(os.path.join(self.path2database, xml_directory, xml_file_name), parser)

                # check if the xml file was already parsed
                dest_file_name = os.path.join(self.path2database, self.path2pub2ref,'{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
                if not rewrite_existing and os.path.isfile(dest_file_name):
                    ifile+=1
                    continue

                pub2ref = []

                all_pubmed_articles = xmltree.findall("/PubmedArticle")

                for article_bucket in all_pubmed_articles:

                    medline = article_bucket.find("MedlineCitation")

                    # scrape the publication information
                    PublicationId = load_int(load_xml_text(medline.find('PMID')))


                    references = article_bucket.find("PubmedData/ReferenceList")
                    if not references is None:
                        for ref in references.findall("Reference"):
                            citation = load_xml_text(ref.find("Citation"))
                            if not ref.find('ArticleIdList') is None:
                                pmid = load_int(load_xml_text(ref.find('ArticleIdList').find('ArticleId[@IdType="pubmed"]')))
                            else:
                                pmid = ""
                            pub2ref.append([PublicationId, pmid, citation])

                # save file
                pub2ref = pd.DataFrame(pub2ref, columns = ['CitedPublicationId', 'CitingPublicationId', 'Citation'], dtype=int)
                fname = os.path.join(self.path2database, self.path2pub2ref, '{}{}.{}'.format(self.path2pub2ref, ifile, self.database_extension))
                self.save_data_file(pub2ref, fname, key =self.path2pub2ref)


        # load the citations into a large dataframe

        pub2ref_files = glob.glob(os.path.join(self.path2database, self.path2pub2ref, '{}*.{}'.format(self.path2pub2ref, self.database_extension)))

        pub2ref = pd.DataFrame()

        print("parsing citation data...")
        for pub2ref_tmp in tqdm(pub2ref_files,desc='PubMed citation xml files', leave=True, disable=not show_progress):
            pub2ref = pub2ref.append(self.read_data_file(pub2ref_tmp), ignore_index=True)

        return pub2ref

    def parse_publicationauthoraffiliation(self, xml_directory = 'RawXML',preprocess = True, num_file_lines=10**7, rewrite_existing = False):
        """
        Parse the PubMed publication-author raw data.
        
        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.
        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.
        num_file_lines: int, default 5*10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.
        show_progress: bool, default True
            Show progress with processing of the data.
       
        Returns
        ----------
        DataFrame
            Publication-Author DataFrame.
        """

        # process author files through xml
        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, self.paa)):
                os.mkdir(os.path.join(self.path2database, self.paa))

            xmlfiles = sorted([fname for fname in os.listdir(os.path.join(self.path2database, xml_directory)) if '.xml' in fname])

            # read dtd - this takes
            path2database = self.path2database # remove self to use inside of this class
            class DTDResolver(etree.Resolver):
                def resolve(self, system_url, public_id, context):
                    return self.resolve_filename(os.path.join(path2database, system_url), context)

            parser = etree.XMLParser(load_dtd=True, resolve_entities=True)

            ifile = 0
            for xml_file_name in tqdm(xmlfiles, desc='PubMed author xml files', leave=True, disable=not show_progress):

                # check if the xml file was already parsed
                dest_file_name = os.path.join(self.path2database, self.path2paa,'{}{}.{}'.format(self.path2paa, ifile, self.database_extension))
                if not rewrite_existing and os.path.isfile(dest_file_name):
                    ifile+=1
                    continue

                paa = []

                all_pubmed_articles = xmltree.findall("/PubmedArticle")

                for article_bucket in all_pubmed_articles:

                    medline = article_bucket.find("MedlineCitation")

                    # scrape the publication information
                    PublicationId = load_int(load_xml_text(medline.find('PMID')))

                    author_list = article.find('AuthorList')

                    if not author_list is None:
                        for seq, author in enumerate(author_list.findall('Author')):
                            author_record = self._blank_pubmed_author()

                            author_record['PublicationId'] = PublicationId
                            author_record['FirstName'] = load_html_str(load_xml_text(author.find("ForeName")))
                            author_record['LastName'] = load_html_str(load_xml_text(author.find("LastName")))
                            author_record['FullName'] = author_record['FirstName'] + ' ' + author_record['LastName']

                            if author.find("AffiliationInfo/Affiliation") is not None:
                                author_record['Affiliations'] = load_html_str(load_xml_text(author.find("AffiliationInfo/Affiliation")))
                                author_record['Affiliations'] = author_record['Affiliations'].replace("For a full list of the authors' affiliations please see the Acknowledgements section.","")

                            author_record['AuthorSequence'] = seq+1

                            paa.append(author_record)
                paa = pd.DataFrame(paa)
                paa['AuthorSequence'] = paa['AuthorSequence'].astype(int)
                self.save_data_file( dest_file_name, key = self.path2paa)
                ifile += 1


        ## load publication author dataframe into a large file
        paa_files_list = glob.glob(os.path.join(self.path2database, self.path2paa, '{}*.{}'.format(self.path2paa, self.database_extension)))

        paa = pd.DataFrame()

        print("Parsing files...")
        for tmp_paa in tqdm(paa_files_list, desc='PubMed author files', leave=True, disable=not show_progress):
            paa = paa.append(self.read_data_file(tmp_paa), ignore_index = True)

        return paa

    def parse_fields(self, preprocess = True, num_file_lines=10**7, rewrite_existing=False,xml_directory = 'RawXML'):
        """
        Parse the PubMed field (mesh term) raw data.
        
        Parameters
        ----------
        preprocess: bool, default True
            Save the processed data in new DataFrames.
        process_name: bool, default True
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.
        num_file_lines: int, default 5*10**6
            The processed data will be saved into smaller DataFrames, each with `num_file_lines` rows.
        show_progress: bool, default True
            Show progress with processing of the data.
        
        Returns
        ----------
        DataFrame
            Publication-Term ID DataFrame and Term ID - Term DataFrame
        """

        if preprocess:
            for hier_dir_type in [self.path2pub2field, self.path2fieldinfo]:
                if not os.path.exists(os.path.join(self.path2database, hier_dir_type)):
                    os.mkdir(os.path.join(self.path2database, hier_dir_type))

            xmlfiles = sorted([fname for fname in os.listdir(os.path.join(self.path2database, xml_directory)) if '.xml' in fname])

            # read dtd - this takes
            path2database = self.path2database # remove self to use inside of this class
            class DTDResolver(etree.Resolver):
                def resolve(self, system_url, public_id, context):
                    return self.resolve_filename(os.path.join(path2database, system_url), context)
            parser = etree.XMLParser(load_dtd=True, resolve_entities=True)

            # global id to term mapping
            fieldinfo = {}

            ifile = 0
            for xml_file_name in tqdm(xmlfiles, desc='PubMed xml files', leave=True, disable=not show_progress):

                # check if the xml file was already parsed
                dest_file_name = os.path.join(self.path2database, self.path2pub2field,'{}{}.{}'.format(self.path2pub2field, ifile, self.database_extension))
                if not rewrite_existing and os.path.isfile(dest_file_name):
                    ifile+=1
                    continue

                pub2field = []
                all_pubmed_articles = xmltree.findall("/PubmedArticle")

                for article_bucket in all_pubmed_articles:

                    medline = article_bucket.find("MedlineCitation")

                    # scrape the publication information
                    PublicationId = load_int(load_xml_text(medline.find('PMID')))

                    meshterms = medline.find("MeshHeadingList")

                    if meshterms is not None:
                        for term in meshterms.getchildren():
                            ui = term.find("DescriptorName").attrib.get("UI", "")
                            if len(ui)>0:
                                pub2field.append([PublicationId, ui])
                                fieldinfo[ui] = [load_xml_text(term.find("DescriptorName")), 'mesh']

                    chemicals = medline.find("ChemicalList")
                    if chemicals is not None:
                        for chemical in chemicals.findall("Chemical"):
                            ui = chemical.find("NameOfSubstance").attrib.get("UI", "")
                            if len(ui)>0:
                                pub2field.append([PublicationId, ui])
                                fieldinfo[ui] = [load_xml_text(chemical.find("NameOfSubstance")), 'chem']

                # save the pub-field id
                pub2field = pd.DataFrame(pub2field, columns = ['PublicationId', 'FieldId'], dtype=int)
                self.save_data_file(dest_file_name, key=self.path2pub2field)

                ifile += 1

            # if rewriting
            dest_file_name = os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format(self.path2fieldinfo, 0, self.database_extension))
            if rewrite_existing:
                # save field info dictionary
                mesh_id_list = list(fieldinfo.values())
                for i, j in enumerate(fieldinfo.keys()):
                    mesh_id_list[i].insert(0, j)

                fieldinfo = pd.DataFrame(mesh_id_list, columns = ['FieldId', 'FieldName', 'FieldType'], dtype=int)
                self.save_data_file(fieldinfo, dest_file_name, key=self.path2fieldinfo)

        # load the dataframes
        # pub2field
        pub2field_files = glob.glob(os.path.join(self.path2database, self.path2pub2field, '{}*.{}'.format(self.path2pub2field, self.database_extension)))
        pub2field = pd.DataFrame()

        for pub2field_tmp_file in tqdm(pub2field_files, desc='PubMed pub2field files', leave=True, disable=not show_progress):
            pub2field = pub2field.append(self.read_data_file(pub2field_tmp_file), ignore_index=True)

        # field info map
        fieldinfo = self.read_data_file(os.path.join(self.path2database, self.path2fieldinfo, '{}{}.{}'.format(self.path2fieldinfo, 0, self.database_extension)))

        return pub2field, fieldinfo



# -*- coding: utf-8 -*-
"""
.. module:: bibnetwork
    :synopsis: The main Network class

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """
import sys
from collections import defaultdict
from itertools import combinations
import numpy as np
import pandas as pd

import scipy.sparse as spsparse

from pyscisci.utils import isin_sorted, zip2dict, check4columns, value_to_int
from pyscisci.sparsenetworkutils import threshold_network, largest_connected_component_vertices, dataframe2bipartite, project_bipartite_mat

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


def coauthorship_network(paa, focus_author_ids=None, focus_constraint='authors', temporal=False, show_progress=False):
    """
    Create the co-authorship network.

    Parameters
    ----------
    paa : DataFrame
        A DataFrame with the links between authors and publications.

    focus_author_ids : numpy array or list, default None
        A list of the AuthorIds to seed the coauthorship-network.

    focus_constraint : str, default 'authors'
        If focus_author_ids is not None:
            - 'authors' : the 'focus_author_ids' defines the node set, giving only the co-authorships between authors in the set.
            - 'publications' : the publication history of `focus_author_ids' defines the edge set, giving the co-authorhips where at least one author from `focus_author_ids' was involved.
            - 'ego' : the 'focus_author_ids' defines a seed set, such that all authors must have co-authored at least one publication with an author from `focus_author_ids', but co-authorships are also found between the second-order author sets.

    temporal : bool, default False
        If True, compute the adjacency matrix using only publications for each year.

    show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

 

    Returns
    -------
    coo_matrix or dict of coo_matrix
        If temporal == False:
            The adjacency matrix for the co-authorship network
        If temporal == True:
            A dictionary with key for each year, and value of the adjacency matrix for the co-authorship network induced by publications in that year.
    author2int, dict
        A mapping of AuthorIds to the row/column of the adjacency matrix.


    |
    

    """
    required_columns = ['AuthorId', 'PublicationId']
    if temporal:
        required_columns.append('Year')
    check4columns(paa, required_columns)
    paa = paa[required_columns].dropna()

    if not focus_author_ids is None:
        focus_author_ids = np.sort(focus_author_ids)

        # identify the subset of the publications we need to form the network
        if focus_constraint == 'authors':
            # take only the publication-author links that have an author from the `focus_author_ids'
            paa = paa.loc[isin_sorted(paa['AuthorId'].values, focus_author_ids)]

        elif focus_constraint == 'publications':
            # take all publications authored by an author from the `focus_author_ids'
            focus_pubs = np.sort(paa.loc[isin_sorted(paa['AuthorId'].values, focus_author_ids)]['PublicationId'].unique())
            # then take only the subset of publication-author links inducded by these publications
            paa = paa.loc[isin_sorted(paa['PublicationId'].values, focus_pubs)]
            del focus_pubs

        elif focus_constraint == 'ego':
            # take all publications authored by an author from the `focus_author_ids'
            focus_pubs = np.sort(paa.loc[isin_sorted(paa['AuthorId'].values, focus_author_ids)]['PublicationId'].unique())
            # then take all authors who contribute to this subset of publications
            focus_author_ids = np.sort(paa.loc[isin_sorted(paa['PublicationId'].values, focus_pubs)]['AuthorId'].unique())
            del focus_pubs
            # finally take the publication-author links that have an author from the above ego subset
            paa = paa.loc[isin_sorted(paa['AuthorId'].values, focus_author_ids)]

    paa.drop_duplicates(subset=['AuthorId', 'PublicationId'], inplace=True)

    #  map authors to the rows of the bipartite adj mat
    author2int = {aid:i for i, aid in enumerate(np.sort(paa['AuthorId'].unique()))}
    Nauthors = paa['AuthorId'].nunique()

    paa['AuthorId'] = [author2int[aid] for aid in paa['AuthorId'].values]

    #  map publications to the columns of the bipartite adj mat
    pub2int = {pid:i for i, pid in enumerate(np.sort(paa['PublicationId'].unique()))}
    Npubs = paa['PublicationId'].nunique()

    paa['PublicationId'] = [pub2int[pid] for pid in paa['PublicationId'].values]

    if temporal:
        years = np.sort(paa['Year'].unique())

        temporal_adj = {}
        for y in years:
            bipartite_adj = dataframe2bipartite(paa.loc[paa['Year'] == y], 'AuthorId', 'PublicationId', (Nauthors, Npubs) )

            adj_mat = project_bipartite_mat(bipartite_adj, project_to = 'row')

            # remove diagonal entries
            adj_mat.setdiag(0)
            adj_mat.eliminate_zeros()

            temporal_adj[y] = adj_mat

        return temporal_adj, author2int

    else:
        bipartite_adj = dataframe2bipartite(paa, 'AuthorId', 'PublicationId', (Nauthors, Npubs) )

        adj_mat = project_bipartite_mat(bipartite_adj, project_to = 'row')

        # remove diagonal entries
        adj_mat.setdiag(0)
        adj_mat.eliminate_zeros()

        return adj_mat, author2int


def extract_multiscale_backbone(Xs, alpha):
    """
    A sparse matrix implemntation of the multiscale backbone.

    References
    ----------
    Serrano et al. (2009) Extracting the multiscale backbone of complex weighted networks.  PNAS.

    Parameters
    ----------
    Xs : numpy.array or sp.sparse matrix
        The adjacency matrix for the network.

    alpha : float
        The significance value.


    Returns
    -------
    coo_matrix
        The directed, weighted multiscale backbone



    |
    

    """

    X = spsparse.coo_matrix(Xs)
    X.eliminate_zeros()

    #normalize
    row_sums = X.sum(axis = 1)
    degrees = X.getnnz(axis = 1)


    pijs = np.multiply(X.data, 1.0/np.array(row_sums[X.row]).squeeze())
    powers = degrees[X.row.squeeze()] - 1

    # equation 2 => where 1 - (k - 1) * integrate.quad(lambda x: (1 - x) ** (k - 2)) = (1-x)**(k - 1) if k  > 1
    significance = np.logical_and(pijs < 1, np.power(1.0 - pijs, powers) < alpha)

    keep_graph = spsparse.coo_matrix((X.data[significance], (X.row[significance], X.col[significance])), shape = X.shape)
    keep_graph.eliminate_zeros()

    return keep_graph


def cocitation_network(pub2ref, focus_pub_ids=None, focus_constraint='citing', cited_col_name = 'CitedPublicationId', 
    citing_col_name = 'CitingPublicationId', temporal=False, show_progress=False):
    """
    Create the co-citation network.


    Parameters
    ----------
    pub2ref : DataFrame
        A DataFrame with the links between authors and publications.

    focus_pub_ids : numpy array or list, default None
        A list of the PublicationIds to seed the cocitation-network.

    focus_constraint : str, default 'citing'
        If focus_author_ids is not None
            -'citing' : the 'focus_pub_ids' defines the citation set, giving only the co-citations between the references
                of the publications from this set.
            -'cited' : the 'focus_pub_ids' defines the cocitation node set.
            -'egocited' : the 'focus_pub_ids' defines a seed set, such that all other publications must have been co-citeed with
                at least one publication from this set.
    
    cited_col_name : str, default 'CitedPublicationId'
        The name of the cited value column in the DataFrame pub2ref
    
    citing_col_name : str, default 'CitingPublicationId'
        The name of the citing value column in the DataFrame pub2ref

    temporal : bool, default False
        If True, compute the adjacency matrix using only publications for each year.

    show_progress : bool, default False
        If True, show a progress bar tracking the calculation.



    Returns
    -------
    coo_matrix or dict of coo_matrix
        If temporal == False:
            The adjacency matrix for the co-citation network

        If temporal == True:
            A dictionary with key for each year, and value of the adjacency matrix for the cocitation network induced
            by citing publications in that year.

    pub2int, dict
        A mapping of PublicationIds to the row/column of the adjacency matrix.

    """
    required_columns = [cited_col_name, citing_col_name]
    if temporal:
        required_columns.append('CitingYear')
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].dropna()

    if not focus_pub_ids is None:
        focus_pub_ids = np.sort(focus_pub_ids)

        # identify the subset of the publications we need to form the network
        if focus_constraint == 'citing':
            # take only the links that have a citing publication from the `focus_pub_ids'
            pub2ref = pub2ref.loc[isin_sorted(pub2ref[citing_col_name].values, focus_pub_ids)]

        elif focus_constraint == 'cited':
            # take only the links that have a cited publication from the `focus_pub_ids'
            pub2ref = pub2ref.loc[isin_sorted(pub2ref[cited_col_name].values, focus_pub_ids)]

        elif focus_constraint == 'egocited':
            # take all publications that cite one of the publications in `focus_pub_ids'
            focus_citing_pubs = np.sort(pub2ref.loc[isin_sorted(pub2ref[cited_col_name].values, focus_pub_ids)][citing_col_name].unique())
            # then take all the links that have a citing publication from the `focus_citing_pubs'
            pub2ref = pub2ref.loc[isin_sorted(pub2ref[citing_col_name].values, focus_citing_pubs)]
            del focus_citing_pubs

    pub2ref.drop_duplicates(subset=[citing_col_name, cited_col_name], inplace=True)

    if pub2ref.shape[0] > 0:
        #  map cited publications to the rows of the bipartite adj mat
        pub2ref[cited_col_name], cited2int = value_to_int(pub2ref[cited_col_name].values, sort_values='value', return_map=True)
        Ncited = len(cited2int)

        #  map citing publications to the columns of the bipartite adj mat
        pub2ref[citing_col_name], citing2int = value_to_int(pub2ref[citing_col_name].values, sort_values='value', return_map=True)
        Nciting = len(citing2int)

        if temporal:
            years = np.sort(pub2ref['CitingYear'].unique())

            temporal_adj = {}
            for y in years:
                bipartite_adj = dataframe2bipartite(pub2ref.loc[pub2ref['CitingYear'] == y], cited_col_name, citing_col_name, (Ncited, Nciting) )

                adj_mat = project_bipartite_mat(bipartite_adj, project_to = 'row')

                # remove diagonal entries
                adj_mat.setdiag(0)
                adj_mat.eliminate_zeros()

                temporal_adj[y] = adj_mat

            return temporal_adj, cited2int

        else:
            bipartite_adj = dataframe2bipartite(pub2ref, cited_col_name, citing_col_name, (Ncited, Nciting) )

            adj_mat = project_bipartite_mat(bipartite_adj, project_to = 'row')

            # remove diagonal entries
            adj_mat.setdiag(0)
            adj_mat.eliminate_zeros()

            return adj_mat, cited2int

    else:
        return spsparse.coo_matrix(), {}

def cociting_network(pub2ref, focus_pub_ids=None, focus_constraint='citing', cited_col_name = 'CitedPublicationId', 
    citing_col_name = 'CitingPublicationId', temporal=False, show_progress=False):
    """
    Create the co-citing network.  Each node is a publication, two publications are linked if they cite the same article.


    Parameters
    ----------

    pub2ref : DataFrame
        A DataFrame with the links between authors and publications.

    focus_pub_ids : numpy array or list, default None
        A list of the PublicationIds to seed the cocitation-network.

    focus_constraint : str, default 'citing'
        If focus_author_ids is not None
            - 'citing' : the 'focus_pub_ids' defines the citation set, giving only the co-citations between the references
                of the publications from this set.
            - 'cited' : the 'focus_pub_ids' defines the cocitation node set.

    cited_col_name : str, default 'CitedPublicationId'
        The name of the cited value column in the DataFrame pub2ref
    
    citing_col_name : str, default 'CitingPublicationId'
        The name of the citing value column in the DataFrame pub2ref

    show_progress : bool, default False
        If True, show a progress bar tracking the calculation.


    Returns
    -------

    coo_matrix or dict of coo_matrix
        The adjacency matrix for the co-citing network

    pub2int, dict
        A mapping of PublicationIds to the row/column of the adjacency matrix.



    |
    

    """
    required_columns = [cited_col_name, citing_col_name]
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].dropna()

    if not focus_pub_ids is None:
        focus_pub_ids = np.sort(focus_pub_ids)

        # identify the subset of the publications we need to form the network
        if focus_constraint == 'citing':
            # take only the links that have a citing publication from the `focus_pub_ids'
            pub2ref = pub2ref.loc[isin_sorted(pub2ref[citing_col_name].values, focus_pub_ids)]

        elif focus_constraint == 'cited':
            # take only the links that have a cited publication from the `focus_pub_ids'
            pub2ref = pub2ref.loc[isin_sorted(pub2ref[cited_col_name].values, focus_pub_ids)]

    pub2ref.drop_duplicates(subset=[citing_col_name, cited_col_name], inplace=True)

    if pub2ref.shape[0] > 0:
        #  map cited publications to the rows of the bipartite adj mat
        pub2ref[cited_col_name], cited2int = value_to_int(pub2ref[cited_col_name].values, sort_values='value', return_map=True)
        Ncited = len(cited2int)

        #  map citing publications to the columns of the bipartite adj mat
        pub2ref[citing_col_name], citing2int = value_to_int(pub2ref[citing_col_name].values, sort_values='value', return_map=True)
        Nciting = len(citing2int)

        bipartite_adj = dataframe2bipartite(pub2ref, cited_col_name, citing_col_name, (Ncited, Nciting) )

        adj_mat = project_bipartite_mat(bipartite_adj, project_to = 'col')

        # remove diagonal entries
        adj_mat.setdiag(0)
        adj_mat.eliminate_zeros()

        return adj_mat, citing2int

    else:
        return spsparse.coo_matrix(), {}

def temporal_cocited_edgedict(pub2ref, pub2year):

    required_pub2ref_columns = ['CitingPublicationId', 'CitedPublicationId']
    check4columns(pub2ref, required_pub2ref_columns)
    pub2ref = pub2ref[required_pub2ref_columns]

    year_values = sorted(list(set(pub2year.values())))

    # we need the citation counts and cocitation network
    temporal_cocitation_dict = {y:defaultdict(set) for y in year_values}
    temporal_citation_dict = {y:defaultdict(int) for y in year_values}

    def count_cocite(cited):
        y = pub2year[cited.name]

        for citedpid in cited['CitedPublicationId'].values:
            temporal_citation_dict[y][citedpid] += 1
        for icitedpid, jcitedpid in combinations(cited['CitedPublicationId'].values, 2):
            temporal_cocitation_dict[y][icitedpid].add(jcitedpid)
            temporal_cocitation_dict[y][jcitedpid].add(icitedpid)

    pub2ref.groupby('CitingPublicationId', sort=False).apply(count_cocite)


def create_citation_edgelist(database, publication_subset = [], temporal = True):

    if len(publication_subset) == 0:
        publication_subset = database.publication_ids()

    edgelist = [[pubid, refid] for pubid in publication_subset for refid in database.get_pub(pubid).references]
    return edgelist

def estimate_resolution(G, com):
    """
    Newman, MEJ (2016) Community detection in networks: Modularity optimization and maximum likelihood are equivalent. Phy. Rev. E
    """
    m = G.number_of_edges()
    
    # eq 16
    kappas = [sum(deg for n, deg in G.degree(c)) for c in com.communities]
    m_in = [G.subgraph(c).number_of_edges() for c in com.communities]
    
    denom = np.sum(np.square(kappas)) / (2*m)
    
    # eq 17
    omega_in = 2*sum(m_in) / denom
    
    # eq 18
    omega_out = (2*m - 2*sum(m_in)) / (2*m - denom)
    
    # eq 15
    gamma = (omega_in - omega_out) / (np.log(omega_in) - np.log(omega_out))
    
    return gamma




# -*- coding: utf-8 -*-
"""
.. module:: pivotscore
    :synopsis: Set of functions for typical bibliometric citation analysis

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """
import sys
import pandas as pd
import numpy as np

from pyscisci.utils import groupby_count, changepoint, pandas_cosine_similarity

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm




### Pivot Score
def pivot_score(pub2author, pub2ref, previous_k=None, year_window=None, show_progress=False):
    """
    Calculate the pivot score index as proposed in :cite:`Hill2022pivotpenalty`.

    None is returned for the first publication in an author's career (because there is no previous history), or when
    no other publications have been published within the year_window.

    Parameters
    ----------
    pub2author : dataframe
        The publication author linkages with Year.

    pub2ref : dataframe
        The citing-cited publication linkages which contains the CitedJournalId for the cited articles.

    previous_k : int, default None
        Only compare against the previous k publications.

    year_window : int, default None
        Only compare against publications from the last year_window years.

    show_progress : bool, default False
        Show calculation progress. 

    Returns
    ----------
    pivotscore : DataFrame
        The pivotscore for each Author Publication.

    """
    if show_progress:
        print("Starting computation of pivot score.")

    pub2refjournalcounts = pyscisci.groupby_count(pub2ref, ['CitingPublicationId', 'CitedJournalId'], 
                                              'CitedPublicationId', count_unique=True)
    pub2refjournalcounts.rename(columns={'CitedPublicationIdCount':'CitedJournalCount'}, inplace=True)
    
    pa_refs = paa.merge(pub2refjournalcounts, how='left', left_on = 'PublicationId', right_on='CitingPublicationId')
    del pa_refs['CitingPublicationId']
    
    pa_refs.dropna(inplace=True)
    pa_refs['CitedJournalId'] = pa_refs['CitedJournalId'].astype(int)
    pa_refs.sort_values(by=['AuthorId', 'Year', 'PublicationId', 'CitedJournalId'], inplace=True)
    pa_refs.reset_index(drop=True, inplace=True)

    pscore = pa_refs.groupby('AuthorId').apply(author_pivot, 
                                               previous_k=previous_k, year_window=year_window).reset_index()
    del pscore['level_1']

    return pscore


def author_pivot(authordf):
        
        pubgroups = authordf.groupby('PublicationId', sort=False)
        
        allpubidx = None
        if not previous_k is None:
            allpubidx = pyscisci.changepoint(authordf['PublicationId'].values)
        
        
        pivotresults = []
        
        def publication_pivot(pubgroup):
            pubidx = pubgroup.index[0]
            pid = pubgroup.name
            if pubidx==0: pivotresults.append([pid, None])
            else:
                i=len(pivotresults)
                if not previous_k is None and i > previous_k:
                    history = authordf.iloc[allpubidx[i-previous_k]:pubidx]
                else:
                    history = authordf.iloc[:pubidx]

                if not year_window is None:
                    history = history[history['Year'] >= pubgroup['Year'].values[0] - year_window]

                if history.shape[0] > 0:
                    history = history.groupby('CitedJournalId', sort=False, as_index=False)['CitedJournalCount'].sum()
                    
                    cosine = pandas_cosine_similarity(history, pubgroup, col_key='CitedJournalId', col_values='CitedJournalCount')
                    
                    pivotresults.append([pid, cosine])
                else:
                    pivotresults.append([pid, None])  
        
        pubgroups.apply(publication_pivot)
                
        return pd.DataFrame(pivotresults, columns=['PublicationId', 'PivotScore'])



# -*- coding: utf-8 -*-
"""
.. module:: sparsenetworkutils
    :synopsis: The main Network class

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """
import sys
import numpy as np
import pandas as pd

from scipy import repeat, sqrt, where, square, absolute
import scipy.sparse as spsparse

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

def threshold_network(adj_mat, threshold=0):
    """

    """
    if adj_mat.getformat() != 'coo':
        adj_mat = spsparse.coo_matrix(adj_mat)

    adj_mat.data[adj_mat.data <=threshold] = 0
    adj_mat.eliminate_zeros()

    return adj_mat

def largest_connected_component_vertices(adj_mat):
    """
    """
    n_components, labels = spsparse.csgraph.connected_components(adj_mat)
    comidx, compsizes = np.unique(labels, return_counts=True)

    return np.arange(adj_mat.shape[0])[labels==np.argmax(compsizes)]


def dataframe2bipartite(df, rowname, colname, shape=None, weightname=None):

    if shape is None:
        shape = (int(df[rowname].max()+1), int(df[colname].max()+1) )

    if weightname is None:
        weights = np.ones(df.shape[0], dtype=int)
    else:
        weights = df[weightname].values

    # create a bipartite adj matrix connecting authors to their publications
    bipartite_adj = spsparse.coo_matrix( ( weights,
                                        (df[rowname].values, df[colname].values) ),
                                        shape=shape, dtype=weights.dtype)

    bipartite_adj.sum_duplicates()

    return bipartite_adj

def project_bipartite_mat(bipartite_adj, project_to = 'row'):

    if project_to == 'row':
        adj_mat = bipartite_adj.dot(bipartite_adj.T).tocoo()
    elif project_to == 'col':
        adj_mat = bipartite_adj.T.dot(bipartite_adj).tocoo()

    return adj_mat

def extract_multiscale_backbone(Xs, alpha):
    """
    A sparse matrix implemntation of the multiscale backbone :cite:`Serrano2009backbone`.

    Parameters
    ----------
    :param Xs : numpy.array or sp.sparse matrix
        The adjacency matrix for the network.

    :param alpha : float
        The significance value.


    Returns
    -------
    coo_matrix
        The directed, weighted multiscale backbone

    """

    X = spsparse.coo_matrix(Xs)
    X.eliminate_zeros()

    #normalize
    row_sums = X.sum(axis = 1)
    degrees = X.getnnz(axis = 1)


    pijs = np.multiply(X.data, 1.0/np.array(row_sums[X.row]).squeeze())
    powers = degrees[X.row.squeeze()] - 1

    # equation 2 => where 1 - (k - 1) * integrate.quad(lambda x: (1 - x) ** (k - 2)) = (1-x)**(k - 1) if k  > 1
    significance = np.logical_and(pijs < 1, np.power(1.0 - pijs, powers) < alpha)

    keep_graph = spsparse.coo_matrix((X.data[significance], (X.row[significance], X.col[significance])), shape = X.shape)
    keep_graph.eliminate_zeros()

    return keep_graph

def sparse_pagerank_scipy(adjmat, alpha=0.85, personalization=None, initialization=None,
                   max_iter=100, tol=1.0e-6, dangling=None):
    
    """
    Pagerank for sparse matrices using the power method
    """

    N, _ = adjmat.shape
    assert(adjmat.shape==(N,N))
    
    if N == 0:
        return np.array([])

    out_strength = np.array(adjmat.sum(axis=1)).flatten()
    out_strength[out_strength != 0] = 1.0 / out_strength[out_strength != 0]
    
    Q = spsparse.spdiags(out_strength.T, 0, *adjmat.shape, format='csr')
    adjmat = Q * adjmat

    # initial vector
    if initialization is None:
        x = repeat(1.0 / N, N)
    
    else:
        x = initialization / initialization.sum()

    # Personalization vector
    if personalization is None:
        p = repeat(1.0 / N, N)
    else:
        p = personalization / personalization.sum()

    # Dangling nodes
    if dangling is None:
        dangling_weights = p
    else:
        dangling_weights = dangling / dangling.sum()

    is_dangling = where(out_strength == 0)[0]

    # power iteration: make up to max_iter iterations
    for _ in range(max_iter):
        xlast = x
        x = alpha * (x * adjmat + sum(x[is_dangling]) * dangling_weights) + \
            (1 - alpha) * p
        # check convergence, l1 norm
        err = absolute(x - xlast).sum()
        if err < N * tol:
            return x

    print('power iteration failed to converge in %d iterations.' % max_iter)


def sparse_eigenvector_centrality_scipy(adjmat, max_iter=100, tol=1.0e-6, initialization=None):

    adjmat = spsparse.csr_matrix(adjmat)

    N, _ = adjmat.shape
    assert(adjmat.shape==(N,N))
    
    if N == 0:
        return np.array([])

    # initial vector
    if initialization is None:
        x = repeat(1.0 / N, N)
    
    else:
        x = initialization / initialization.sum()

    
    # make up to max_iter iterations
    for _ in range(max_iter):
        xlast = x
        x = xlast.copy()  # Start with xlast times I to iterate with (A+I)
        # do the multiplication y^T = x^T A (left eigenvector)
        x = xlast*adjmat

        
        norm = sqrt( square(x).sum() ) or 1
        x = x/norm

        # Check for convergence (in the L_1 norm).
        err = absolute(x - xlast).sum()
        if err < N * tol:
            return x
    print('power iteration failed to converge in %d iterations.' % max_iter)



# -*- coding: utf-8 -*-
"""
.. module:: career topics
    :synopsis: Calculate the co-citing network and detect career communities.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np
import networkx as nx
import scipy.sparse as spsparse


# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import isin_sorted
from pyscisci.network import cociting_network
from pyscisci.sparsenetworkutils import largest_connected_component_vertices

## Career Cociting Network
def career_cociting_network_topics(paa, pub2ref, randomize=None, return_network=False, show_progress=False):
    """
    This function calculates the topics throughout a career based on the co-citing network (two publications are linked if they share a reference).  
    See :cite:`zeng2019topicswitch` for details.

    Parameters
    ----------
    paa : dataframe
        The publication author affiliation linkages for the focus author.

    pub2ref : dataframe
        The citing-cited publication linkages which contains the citing articles from the focus author.

    randomize : int, default None
        See for random initialization of the community detection algorithm.

    return_network : bool, default False
        Return the networkx object of the co-citing network.

    show_progress : bool, default False
        Show calculation progress. 

    Returns
    ----------
    switching_career : DataFrame
        The paa with topic number included (topics are detected using the Louvain algorithm in the co-citing network).

    cociting_net : networkx.Graph(), optional
        If 'return_network == True' then the cociting network is returned as a networkx graph object.

    """
    try:
        from cdlib import algorithms
    except ImportError:
        raise ImportError("Optional package cdlib needed for this analysis: pip install cdlib")

    try:
        from clusim.clustering import Clustering
    except ImportError:
        raise ImportError("Optional package clusim needed for this analysis: pip install clusim")

    focus_pub_ids = np.sort(paa['PublicationId'].unique())

    # find the co-citing network
    cociting_adjmat, cociting2int = cociting_network(pub2ref, focus_pub_ids=focus_pub_ids, 
        focus_constraint='citing', 
        cited_col_name = 'CitedPublicationId', 
        citing_col_name = 'CitingPublicationId')

    # now take the largest connected component
    lcc_nodes = largest_connected_component_vertices(cociting_adjmat)

    remapnodes = {nid:i for i, nid in enumerate(lcc_nodes)}
    cociting2int = {pid:remapnodes[i] for pid, i in cociting2int.items() if not remapnodes.get(i, None) is None}

    lcc_cociting_adjmat = spsparse.csr_matrix(cociting_adjmat)[lcc_nodes][:,lcc_nodes]

    # remove self-loops and binarize
    lcc_cociting_adjmat.setdiag(0)
    lcc_cociting_adjmat.data[lcc_cociting_adjmat.data >1] = 1
    lcc_cociting_adjmat.eliminate_zeros()

    lcc_cociting_net = nx.Graph(lcc_cociting_adjmat)
    coms = algorithms.louvain(lcc_cociting_net, resolution=1., randomize=randomize)
    louvain_communities = Clustering().from_cluster_list(coms.communities)

    pub2topiccomm = {pid:list(louvain_communities.elm2clu_dict[pid])[0] for pid in lcc_cociting_net.nodes()}

    switching_career = paa[['PublicationId', 'AuthorId', 'Year']].copy()
    switching_career.drop_duplicates(subset=['PublicationId'], inplace=True)
    switching_career = switching_career.loc[isin_sorted(switching_career['PublicationId'].values, np.sort(list(cociting2int.keys())))]
    
    switching_career['TopicCommunity'] = [pub2topiccomm[cociting2int[pid]] for pid in switching_career['PublicationId'].values] 

    switching_career['Degree'] = [lcc_cociting_net.degree()[cociting2int[pid]] for pid in switching_career['PublicationId'].values] 

    switching_career.dropna(inplace=True)

    switching_career.sort_values('Year', inplace=True)

    if return_network:
        nx.set_node_attributes(lcc_cociting_net, pub2topiccomm, "TopicCommunity")
        nx.set_node_attributes(lcc_cociting_net, {i:pid for pid,i in cociting2int.items()}, "PublicationId")

        return switching_career, lcc_cociting_net
    else:
        return switching_career
    


# -*- coding: utf-8 -*-
"""
.. module:: hindex
    :synopsis: Calculate the hindex.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict


### H index

def hindex(a):
    """
    Calculate the h index for the array of citation values.  See :cite:`hirsch2005index` for the definition.

    Parameters
    ----------
    :param a : numpy array
        An array of citation counts for each publication by the Author.

    Returns
    -------
    int
        The Hindex

    """
    d = np.sort(a)[::-1] - np.arange(a.shape[0])
    return (d>0).sum()

def compute_hindex(df, colgroupby, colcountby, show_progress=False):
    """
    Calculate the h index for each group in the DataFrame.  See :cite:`hirsch2005index` for the definition.

    The algorithmic implementation for each author can be found in :py:func:`citationanalysis.author_hindex`.

    Parameters
    ----------
    :param df : DataFrame
        A DataFrame with the citation information for each Author.

    :param colgroupby : str
        The DataFrame column with Author Ids.

    :param colcountby : str
        The DataFrame column with Citation counts for each publication.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: colgroupby, 'Hindex'

        """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Hindex', disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colgroupby)+'Hindex']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].progress_apply(hindex).to_frame().reset_index().rename(columns=newname_dict)

def gindex(a):
    """
    Calculate the g index for the array of citation values.  See :cite:`Waltman2008index` for detailed definition.

    Parameters
    ----------
    :param a : numpy array
        An array of citation counts for each publication by the Author.

    Returns
    -------
    int
        The Gindex

    """
    d = np.cumsum(np.sort(a)[::-1]) - np.arange(a.shape[0])**2
    return (d>0).sum()

def compute_gindex(df, colgroupby, colcountby, show_progress=False):
    """
    Calculate the g index for each group in the DataFrame.  See :cite:`Waltman2008index` for detailed definition.

    The algorithmic implementation for each author can be found in :py:func:`citationanalysis.author_gindex`.

    Parameters
    ----------
    :param df : DataFrame
        A DataFrame with the citation information for each Author.

    :param colgroupby : str
        The DataFrame column with Author Ids.

    :param colcountby : str
        The DataFrame column with Citation counts for each publication.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: colgroupby, 'Hindex'

        """
    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Gindex', disable= not show_progress)

    newname_dict = zip2dict([str(colcountby), '0'], [str(colgroupby)+'Gindex']*2)
    return df.groupby(colgroupby, sort=False)[colcountby].progress_apply(gindex).to_frame().reset_index().rename(columns=newname_dict)



# -*- coding: utf-8 -*-
"""
.. module:: datasource
    :synopsis: Set of classes to work with different bibliometric data sources

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import os
import errno
import json
import gzip
from collections import defaultdict

import pandas as pd
import numpy as np

from dask.distributed import Client
import dask.dataframe as dd

from pyscisci.utils import isin_sorted, zip2dict, groupby_count
#from pyscisci.methods import *
from pyscisci.datasource.readwrite import load_preprocessed_data, append_to_preprocessed
from pyscisci.filter import *

class BibDataBase(object):
    """
    Base class for all bibliometric database interfaces.

    The BibDataBase provides a parasomonious structure for each of the specific data sources (MAG, WOS, etc.).

    There are four primary types of functions:
        1. *Parseing* Functions (data source specific) that parse the raw data files
        2. *Loading* Functions that load DataFrames
        3. *Processing* Functions that calculate advanced data types from the raw files
        4. *Analysis* Functions that calculate Science of Science metrics.

    |

    Parameters
    -------

    path2database: str
        The path to the database files

    keep_in_memory: bool, default False
        Flag to keep database files in memory once loaded

    global_filter: dict or Filter
        Set of conditions to apply globally.

    show_progress: bool, default True
        Flag to show progress of loading, processing, and calculations.
    

    |


    """

    def __init__(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, global_filter = None, show_progress=True):

        self._default_init(path2database, database_extension, keep_in_memory, global_filter, show_progress)
        

    def _default_init(self, path2database = '', database_extension='csv.gz', keep_in_memory = False, 
        global_filter = None, enable_dask=False, show_progress=True):

        self.path2database = path2database
        self.keep_in_memory = keep_in_memory
        self.global_filter = None
        self.use_dask = enable_dask
        self.show_progress = show_progress

        self.path2author = 'author'
        self.path2pub = 'publication'
        self.path2affiliation = 'affiliation'
        self.path2pub2ref = 'pub2ref'
        self.path2paa = 'publicationauthoraffiliation'
        self.path2pub2field = 'pub2field'
        self.path2journal = 'journal'
        self.path2fieldinfo = 'fieldinfo'
        self.path2impact = 'impact'
        self.path2pub2abstract = 'pub2abstract'

        self.database_extension = database_extension

        self._affiliation = None
        self._pub = None
        self._journal = None
        self._author = None
        self._pub2year = None
        self._pub2doctype = None
        self._pub2ref = None
        self._author2pub = None
        self._paa = None
        self._pub2refnoself = None
        self._pub2field=None
        self._fieldinfo = None

        self.PublicationIdType = int
        self.AffiliationIdType = int
        self.AuthorIdType = int
        self.JournalIdType = int

        if not global_filter is None:
            self.set_global_filters(global_filter)

    def set_new_data_path(self, dataframe_name='', new_path=''):
        """
        Override path to the dataframe collection.

        Parameters
        --------
        dataframe_name : str
            The dataframe name to override.  E.g. 'author', 'pub', 'paa', 'pub2field', etc.

        new_path : str
            The new dataframe path.

        """
        
        setattr(self, 'path2' + dataframe_name, new_path)

    def setup_dask_client(self, n_workers=None, threads_per_worker=None):
        """
        Initialize a Dask client

        Parameters
        --------
        n_workers : int, defualt None
            The number of workers to use.

        threads_per_worker : int, defualt None
            The number of threads per worker to use.

        """
        self.dask_client = Client(n_workers=n_workers, threads_per_worker=threads_per_worker)

    @property
    def affiliation(self):
        """
        The DataFrame keeping affiliation information. Columns may depend on the specific datasource.

        Notes
        --------
        columns: 'AffiliationId', 'NumberPublications', 'NumberCitations', 'FullName', 'GridId', 'OfficialPage', 'WikiPage', 'Latitude', 'Longitude'

        
        |

        """
        if self._affiliation is None:
            if self.keep_in_memory:
                self._affiliation = self.load_affiliations(show_progress=self.show_progress)
            else:
                return self.load_affiliations(show_progress=self.show_progress)

        return self._affiliation


    @property
    def author(self):
        """
        The DataFrame keeping author information. Columns may depend on the specific datasource.

        Notes
        --------
        columns: 'AuthorId', 'LastKnownAffiliationId', 'NumberPublications', 'NumberCitations', 'FullName', 'LastName', 'FirstName', 'MiddleName'


        |

        """
        if self._author is None:
            if self.keep_in_memory:
                self._author = self.load_authors(show_progress=self.show_progress)
            else:
                return self.load_authors(show_progress=self.show_progress)

        return self._author

    @property
    def pub(self):
        """
        The DataFrame keeping publication information. Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'PublicationId', 'Year', 'JournalId', 'FamilyId', 'Doi', 'Title', 'Date', 'Volume', 'Issue', 'DocType'

        
        |

        """
        if self._pub is None:
            if self.keep_in_memory:
                self._pub = self.load_publications(show_progress=self.show_progress)
            else:
                return self.load_publications(show_progress=self.show_progress)

        return self._pub

    @property
    def pub2year(self):
        """
        A dictionary mapping PublicationId to Year.

        """
        if self._pub2year is None:
            if self.keep_in_memory:
                self._pub2year = self.load_pub2year()
            else:
                return self.load_pub2year()

        return self._pub2year

    @property
    def pub2doctype(self):
        """
        A dictionary mapping PublicationId to Document Type.
        
        Notes
        -------
        doctype mapping: 'Journal': 'j', 'Book':'b', '':'', 'BookChapter':'bc', 'Conference':'c', 'Dataset':'d', 'Patent':'p', 'Repository':'r'
        
        |

        """
        if self._pub2doctype is None:
            if self.keep_in_memory:
                self._pub2doctype = self.load_pub2doctype()
            else:
                return self.load_pub2doctype()

        return self._pub2doctype

    @property
    def journal(self):
        """
        The DataFrame keeping journal information.  Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'JournalId', 'FullName', 'Issn', 'Publisher', 'Webpage'
        
        |

        """
        if self._journal is None:
            if self.keep_in_memory:
                self._journal = self.load_journals(show_progress=self.show_progress)
            else:
                return self.load_journals(show_progress=self.show_progress)

        return self._journal

    @property
    def pub2ref(self):
        """
        The DataFrame keeping citing and cited PublicationId.

        Notes
        -------
        columns: 'CitingPublicationId', 'CitedPublicationId'
        
        |

        """
        if self._pub2ref is None:
            if self.keep_in_memory:
                self._pub2ref = self.load_references(show_progress=self.show_progress)
            else:
                return self.load_references(show_progress=self.show_progress)

        return self._pub2ref

    @property
    def pub2refnoself(self):
        """
        The DataFrame keeping citing and cited PublicationId after filtering out the self-citations.

        Notes
        -------
        columns: CitingPublicationId, CitedPublicationId
        
        |

        """
        if self._pub2refnoself is None:
            if self.keep_in_memory:
                self._pub2refnoself = self.load_references(noselfcite=True, show_progress=self.show_progress)
            else:
                return self.load_references(noselfcite=True, show_progress=self.show_progress)

        return self._pub2refnoself

    @property
    def paa(self):
        """
        The DataFrame keeping all publication, author, affiliation relationships.  Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'PublicationId', 'AuthorId', 'AffiliationId', 'AuthorSequence',  'OrigAuthorName', 'OrigAffiliationName'
        
        |

        """
        if self._paa is None:
            if self.keep_in_memory:
                self._paa = self.load_publicationauthoraffiliation(show_progress=self.show_progress)
            else:
                return self.load_publicationauthoraffiliation(show_progress=self.show_progress)

        return self._paa

    @property
    def author2pub(self):
        """
        The DataFrame keeping all publication, author relationships.  Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'PublicationId', 'AuthorId'
        
        |

        """
        if self._paa is None:
            if self.keep_in_memory:
                self._paa = self.load_publicationauthoraffiliation(columns = ['AuthorId', 'PublicationId'], database_extension=self.database_extension,
                    duplicate_subset = ['AuthorId', 'PublicationId'], dropna = ['AuthorId', 'PublicationId'], show_progress=self.show_progress)
            else:
                return self.load_publicationauthoraffiliation(columns = ['AuthorId', 'PublicationId'], database_extension=self.database_extension,
                    duplicate_subset = ['AuthorId', 'PublicationId'], dropna = ['AuthorId', 'PublicationId'], show_progress=self.show_progress)

        return self._paa

    @property
    def pub2field(self):
        """
        The DataFrame keeping all publication field relationships.  Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'PublicationId', 'FieldId'
        
        |

        """

        if self._pub2field is None:
            if self.keep_in_memory:
                self._pub2field = self.load_pub2field(show_progress=self.show_progress)
            else:
                return self.load_pub2field(show_progress=self.show_progress)

        return self._pub2field

    @property
    def fieldinfo(self):
        """
        The DataFrame keeping all publication field relationships.  Columns may depend on the specific datasource.

        Notes
        -------
        columns: 'FieldId', 'FieldLevel', 'NumberPublications', 'FieldName'
        
        |

        """
        if self._fieldinfo is None:
            if self.keep_in_memory:
                self._fieldinfo = self.load_fieldinfo(show_progress=self.show_progress)
            else:
                return self.load_fieldinfo(show_progress=self.show_progress)

        return self._fieldinfo

    def publicationid_list(self):
        """
        A list of all PublicationIds.
        

        |

        """
        if self.global_filter is None:
            return sorted(list(self.load_publications(columns=['PublicationId'], duplicate_subset=['PublicationId'])['PublicationId'].unique()))

        else:
            return sorted(list(self.global_filter))


    def set_global_filters(self, global_filter):
        """
        Set of filtering conditions that restrict the global set of publications loaded from the DataBase.

        Allowable global filters are:
            'Year', 'DocType', 'FieldId'
        

        |

        """

        if isinstance(global_filter, list):
            filter_dict = {f.field:f for f in global_filter}

        elif isinstance(global_filter, (RangeFilter, YearFilter, SetFilter, DocTypeFilter, FieldFilter) ):
            filter_dict = { global_filter.field:global_filter}

        elif type(global_filter) is dict:
            filter_dict = global_filter

        else:
            raise TypeError("global_filter must be a single pyscisci Filter, a list of pyscisci Filters, or a dictionary of pyscisci Filters.")

        # to create a global filter across all dataframes, we need to create a list of the allowable PublicationIds
        
        for filtertype, pubfilter in filter_dict.items():
            if filtertype == 'Year':
                pub2year = self.pub2year

                if self.global_filter is None:
                    self.global_filter = {pid for pid, y in pub2year.items() if filter_dict['Year'].check_value(y)}
                else:
                    self.global_filter = {pid for pid in self.global_filter if not pub2year.get(pid, None) is None and filter_dict['Year'].check_value(pub2year[pid])}

            elif filtertype == 'DocType':
                pub2doctype = self.pub2doctype
                
                if self.global_filter is None:
                    self.global_filter = {pid for pid, dt in pub2doctype.items() if filter_dict['DocType'].check_value(dt)}
                else:
                    self.global_filter = {pid for pid in self.global_filter if not pub2doctype.get(pid, None) is None and filter_dict['DocType'].check_value(pub2doctype[pid])}

            elif filtertype == 'FieldId':

                if self.global_filter is None:
                    load_fields_filter = {'FieldId':np.sort(list(filter_dict['FieldId'].value_set))}

                else:
                    load_fields_filter = {'FieldId':np.sort(list(filter_dict['FieldId'].value_set)), 'PublicationId':np.sort(list(self.global_filter))}

                pub2field = self.load_pub2field(columns = ['PublicationId', 'FieldId'], filter_dict = load_fields_filter)

                self.global_filter = set(pub2field['PublicationId'].unique())

            else:
                raise TypeError("{} is not a valid global filter.  It must be one of 'Year', 'DocType', 'FieldId'.".format(filtertype))


    ## Basic Functions for loading data from either preprocessed sources or the raw database files

    def load_affiliations(self, preprocess = True, columns = None, filter_dict = None, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the Affiliation DataFrame from a preprocessed directory, or parse from the raw files.
        

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default None, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns
        
   
        Returns
        -------
        DataFrame
            Affililation DataFrame.
        

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Affiliations'

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2affiliation)):
                return load_preprocessed_data(self.path2affiliation, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2affiliation))
        else:
            return self.parse_affiliations()

    def load_authors(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, process_name = True, show_progress=True):
        """
        Load the Author DataFrame from a preprocessed directory, or parse from the raw files.


        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        process_name : bool, default True, Optional
            If True, then when processing the raw file, the package `NameParser <https://nameparser.readthedocs.io/en/latest/>`_
            will be used to split author FullNames.


        Returns
        -------
        DataFrame
            Author DataFrame.
        

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Authors'

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2author)):
                return load_preprocessed_data(self.path2author, path2database=self.path2database, database_extension=self.database_extension, 
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2author))
        else:
            return self.parse_authors(process_name=process_name)

    def load_publications(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the Publication DataFrame from a preprocessed directory, or parse from the raw files.


        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns


        Returns
        -------
        DataFrame
            Publication DataFrame.


        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Publications'

        if not self.global_filter is None:
            if not filter_dict is None and 'PublicationId' in filter_dict:
                filter_dict['PublicationId'] = np.sort([pid for pid in filter_dict['PublicationId'] if pid in self.global_filter])
            else:
                filter_dict = {'PublicationId':np.sort(list(self.global_filter))}

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2pub)):
                return load_preprocessed_data(dataname=self.path2pub, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, use_dask=self.use_dask, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2pub))
        else:
            return self.parse_publications()

    def load_pub2year(self):

        if os.path.exists(os.path.join(self.path2database, 'pub2year.json.gz')):
            with gzip.open(os.path.join(self.path2database, 'pub2year.json.gz'), 'r') as infile:
                pub2year = json.loads(infile.read().decode('utf8'))
            return {self.PublicationIdType(k):int(y) for k,y in pub2year.items() if not y is None}
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), 'pub2year.json.gz')

    def load_pub2doctype(self):
        if os.path.exists(os.path.join(self.path2database, 'pub2doctype.json.gz')):
            with gzip.open(os.path.join(self.path2database, 'pub2doctype.json.gz'), 'r') as infile:
                pub2doctype = json.loads(infile.read().decode('utf8'))
            return {self.PublicationIdType(k):dt for k,dt in pub2doctype.items() if not dt is None}
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), 'pub2doctype.json.gz')

    def load_journals(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the Journal DataFrame from a preprocessed directory, or parse from the raw files.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default None, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        Returns
        -------
        DataFrame
            Journal DataFrame.
        

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Journals'

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2journal)):
                return load_preprocessed_data(self.path2journal, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2journal))
        else:
            return self.parse_publications()

    def load_references(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', noselfcite = False, dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the Pub2Ref DataFrame from a preprocessed directory, or parse from the raw files.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        noselfcite : bool, default False, Optional
            If True, then the preprocessed pub2ref files with self-citations removed will be used.

        Returns
        -------
        DataFrame
            Pub2Ref DataFrame.
        

        |

        """
        if noselfcite:
            fileprefix = self.path2pub2ref + 'noself'
        else:
            fileprefix = self.path2pub2ref

        if filter_dict is None:
            filter_dict = {}

        if not self.global_filter is None:
            if 'CitingPublicationId' in filter_dict:
                filter_dict['CitingPublicationId'] = np.sort([pid for pid in filter_dict['CitingPublicationId'] if pid in self.global_filter])
            else:
                filter_dict['CitingPublicationId'] = np.sort(list(self.global_filter))

            if 'CitedPublicationId' in filter_dict:
                filter_dict['CitedPublicationId'] = np.sort([pid for pid in filter_dict['CitedPublicationId'] if pid in self.global_filter])
            else:
                filter_dict['CitedPublicationId'] = np.sort(list(self.global_filter))

        if show_progress or self.show_progress:
            show_progress='Loading {}'.format(fileprefix)

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, fileprefix)):
                return load_preprocessed_data(fileprefix, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, fileprefix))
        else:
            return self.parse_references()

    def load_publicationauthoraffiliation(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the PublicationAuthorAffilation DataFrame from a preprocessed directory, or parse from the raw files.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        Returns
        -------
        DataFrame
            PublicationAuthorAffilation DataFrame.
        

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Publication Author Affiliation'

        if filter_dict is None:
            filter_dict = {}

        if not self.global_filter is None:
            if 'PublicationId' in filter_dict:
                filter_dict['PublicationId'] = np.sort([pid for pid in filter_dict['PublicationId'] if pid in self.global_filter])
            else:
                filter_dict['PublicationId'] = np.sort(list(self.global_filter))

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2paa)):
                return load_preprocessed_data(self.path2paa, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2paa))
        else:
            return self.parse_publicationauthoraffiliation()

    def load_pub2field(self, preprocess = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the Pub2Field DataFrame from a preprocessed directory, or parse from the raw files.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        Returns
        -------
        DataFrame
            Pub2Field DataFrame.
        

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Fields'

        if filter_dict is None:
            filter_dict = {}

        if not self.global_filter is None:
            if 'PublicationId' in filter_dict:
                filter_dict['PublicationId'] = np.sort([pid for pid in filter_dict['PublicationId'] if pid in self.global_filter])
            else:
                filter_dict['PublicationId'] = np.sort(list(self.global_filter))

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2pub2field)):
                return load_preprocessed_data(self.path2pub2field, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                    prefunc2apply=prefunc2apply, postfunc2apply=postfunc2apply, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2pub2field))
        else:
            return self.parse_fields()

    def load_fieldinfo(self, preprocess = True, columns = None, filter_dict = {}, show_progress=False):
        """
        Load the Field Information DataFrame from a preprocessed directory, or parse from the raw files.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            Attempt to load from the preprocessed directory.

        columns : list, default None, Optional
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        Returns
        -------
        DataFrame
            FieldInformation DataFrame.
    

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Field Info'

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2fieldinfo)):
                return load_preprocessed_data(self.path2fieldinfo, path2database=self.path2database, database_extension=self.database_extension)

            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2fieldinfo))
        else:
            return self.parse_fields()

    def load_impact(self, preprocess = True, include_yearnormed = True, columns = None, filter_dict = {}, duplicate_subset = None,
        duplicate_keep = 'last', dropna = None, prefunc2apply=None, postfunc2apply=None, show_progress=False):
        """
        Load the precomputed impact DataFrame from a preprocessed directory.

        Parameters
        ----------
        preprocess : bool, default True
            Attempt to load from the preprocessed directory.

        include_yearnormed: bool, default True
            Normalize all columns by yearly average.

        columns : list, default None
            Load only this subset of columns

        filter_dict : dict, default {}, Optional
            Dictionary of format {"ColumnName":"ListofValues"} where "ColumnName" is a data column
            and "ListofValues" is a sorted list of valid values.  A DataFrame only containing rows that appear in
            "ListofValues" will be returned.

        duplicate_subset : list, default None, Optional
            Drop any duplicate entries as specified by this subset of columns

        duplicate_keep : str, default 'last', Optional
            If duplicates are being dropped, keep the 'first' or 'last'
            (see `pandas.DataFram.drop_duplicates <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html>`_)

        dropna : list, default None, Optional
            Drop any NaN entries as specified by this subset of columns

        Returns
        -------
        DataFrame
            FieldInformation DataFrame.
        

        |

        """
        if show_progress or self.show_progress:
            show_progress='Loading Impact'

        if filter_dict is None:
            filter_dict = {}

        if not self.global_filter is None:
            if 'PublicationId' in filter_dict:
                filter_dict['PublicationId'] = np.sort([pid for pid in filter_dict['PublicationId'] if pid in self.global_filter])
            else:
                filter_dict['PublicationId'] = np.sort(list(self.global_filter))

        if include_yearnormed:
            def normfunc(impactdf):
                impactcolumns = [c for c in list(impactdf) if not c in ['PublicationId', 'Year']]
                for c in impactcolumns:
                    impactdf[c+'_norm'] = impactdf[c]/impactdf[c].mean()
                return impactdf
        else:
            def normfunc(impactdf):
                return impactdf

        if preprocess:
            if os.path.exists(os.path.join(self.path2database, self.path2impact)):
                return load_preprocessed_data(self.path2impact, path2database=self.path2database, database_extension=self.database_extension,
                    columns=columns, filter_dict=filter_dict, duplicate_subset=duplicate_subset, duplicate_keep=duplicate_keep, dropna=dropna,
                prefunc2apply=normfunc, show_progress=show_progress)
            else:
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2impact))
        else:
            raise self.compute_impact()


    def get_publication_info(pub_ids = [], pub=True, columns = None):
        """
        Produce a DataFrame with the publication information.

        Parameters
        -----------
        pub_ids : int, str, or list
            A PublicationId or list of PublicationIds

        pub : Bool, or DataFrame - default True, optional
            A pre-loaded pub DataFrame or a bool indicating you should load the publicaiton dataframe.

        columns : list, defualt None, optional
            A list of specific columns from the pub DataFrame to return.


        Returns
        -----------
        pub_info : DataFrame
            The publication information.


        """

        if isinstance(pub_ids, self.PublicationIdType):
            pub_ids = [pub_ids]
        pub_ids = np.sort(pub_ids)

        if isinstance(pub, pd.DataFrame) and 'PublicationId' in list(pub):
            pub_info = pub[isin_sorted(pub['PublicationId'].values, pub_ids)]

        elif pub and self._pub is None:
            pub_info = self.load_publications(filter_dict={'PublicationId':pub_ids}, columns=columns)

        elif pub:
            pub_info = self._pub[isin_sorted(self._pub['PublicationId'].values, pub_ids)]

        if not columns is None:
            pub_info = pub_info[columns]

        return pub_info

    def get_author_publications(author_ids = [], paa=None, include_pub_info=None, include_impact=False):
        """
        Produce a DataFrame with the author(s) publication information.

        Parameters
        -----------
        author_ids : int, str, or list
            An AuthorId or list of AuthorIds

        paa : None, or DataFrame - default None, optional
            A pre-loaded paa DataFrame or defualt None indicates we should load the paa dataframe.

        include_pub_info : bool or DataFrame, defualt None, optional
            If DataFrame- used as the publication information.
            If Bool and True - we load our own publication information.

        include_impact : bool or DataFrame, defualt None, optional
            If DataFrame- used as the impact or reference information.
            If Bool and True - we load our own impact information.


        Returns
        -----------
        author_paa : DataFrame
            The author(s) publication information.
        """

        if isinstance(author_ids, self.AuthorIdType):
            author_ids = [author_ids]
        author_ids = np.sort(author_ids)

        if paa is None and self._paa is None:
            author_paa = self.load_publicationauthoraffiliation(columns = ['AuthorId', 'PublicationId', 'AffiliationId'], 
                filter_dict={'AuthorId':author_ids})
        elif paa is None:
            author_paa = self._paa[isin_sorted(self._paa['AuthorId'].values, author_ids)]
        else:
            author_paa = paa[isin_sorted(paa['AuthorId'].values, author_ids)]

        if include_pub_info:
            author_pubs = self.get_publication_info(pub_ids=np.sort(author_paa['PublicationId'].unique()))
            author_paa = author_paa.merge(author_pubs, how='left', on='PublicationId')
            

        return author_paa

    def get_journal_publications(journal_ids = [], pub=None, include_pub_info=None, include_impact=False):
        """
        Produce a DataFrame with the journal(s) publication information.

        Parameters
        -----------
        journal_ids : int, str, or list
            A JournalId or list of JournalIds

        pub : None, or DataFrame - default None, optional
            A pre-loaded pub DataFrame or defualt None indicates we should load the pub dataframe.

        include_pub_info : bool or DataFrame, defualt None, optional
            If DataFrame- used as the publication information.
            If Bool and True - we load our own publication information.

        include_impact : bool or DataFrame, defualt None, optional
            If DataFrame- used as the impact or reference information.
            If Bool and True - we load our own impact information.


        Returns
        -----------
        journal_pubs : DataFrame
            The journal(s) publication information.
        """

        if isinstance(journal_ids, self.JournalIdType):
            journal_ids = [journal_ids]
        journal_ids = np.sort(journal_ids)

        if isinstance(pub, pd.DataFrame) and 'JournalId' in list(pub):
            journal_pubs = pub[isin_sorted(pub['JournalId'].values, journal_ids)]

        elif pub and self._pub is None:
            journal_pubs = self.load_publications(filter_dict={'JournalId':journal_ids}, columns=columns)

        elif pub:
            journal_pubs = self._pub[isin_sorted(self._pub['JournalId'].values, journal_ids)]

        if not columns is None:
            journal_pubs = journal_pubs[columns]

        return journal_pubs


    def publication_citations(pub_ids, imapct=None):
        pass
        

    def save_data_file(self, df, fname, key=''):
        """
        Save the DataFrame to a file.

        Parameters
        -----------
        df : DataFrame
            A pandas DataFrame

        fname : std,
            The filename

        key : str
            For hdf files, the table key.


        Returns
        -----------
        journal_pubs : DataFrame
            The journal(s) publication information.
        """
        if self.database_extension == 'hdf':
            df.to_hdf(fname, mode = 'w', key=key) #, format='table')
        elif self.database_extension == 'csv':
            df.to_csv(fname, mode='w', header=True, index=False, escapechar='\\')
        elif self.database_extension == 'csv.gz':
            df.to_csv(fname, mode='w', header=True, index=False, escapechar='\\', compression='gzip')
        else:
            df.to_csv(fname, mode='w', header=True, index=False, escapechar='\\', compression=self.database_extension)

    def read_data_file(self, fname, key=''):
        """
        Read the DataFrame from a file.

        Parameters
        -----------
        df : DataFrame
            A pandas DataFrame

        fname : std,
            The filename


        Returns
        -----------
        df : DataFrame
            The DataFrame.
        """
        if self.database_extension == 'hdf':
            return pd.read_hdf(fname, key=key)

        elif self.database_extension == 'csv':
            return pd.read_csv(fname, mode='r')

        elif self.database_extension == 'csv.gz':
            return pd.read_csv(fname, mode='r', compression='gzip')

        else:
            return pd.read_csv(fname, mode='r', compression=database_extension)


    """
    To be rewritten for each specific data source (MAG, WOS, etc.)
    """

    def download_from_source(self):
        raise NotImplementedError

    def parse_affiliations(self, preprocess = False):
        raise NotImplementedError

    def parse_authors(self, preprocess = False, process_name = True, num_file_lines = 5*10**6):
        raise NotImplementedError

    def parse_publications(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError

    def parse_references(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError

    def parse_publicationauthoraffiliation(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError

    def parse_fields(self, preprocess = False, num_file_lines=10**7):
        raise NotImplementedError

    def remove_selfcitations(self, preprocess=True, show_progress=False):
        """
        Prcoess the pub2ref DataFrame and remove all citation relationships that share an Author.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            If True then the new preprocessed DataFrames are saved in pub2refnoself

        Returns
        -------
        DataFrame
            Pub2Ref DataFrame with 2 columns: 'CitingPublicationId', 'CitedPublicationId'
        
        |

        """
        if self.show_progress:
            show_progress=True

        fileprefix = self.path2pub2ref + 'noself'

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, fileprefix)):
                os.mkdir(os.path.join(self.path2database, fileprefix))

            if not os.path.exists(os.path.join(self.path2database, self.path2pub2ref)):
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), os.path.join(self.path2database, self.path2pub2ref))

        pub2authors = defaultdict(set)
        for pid, aid in self.author2pub[['PublicationId', 'AuthorId']].values:
            pub2authors[pid].add(aid)

        fullrefdf = []
        # loop through all pub2ref files
        Nreffiles = sum('pub2ref' in fname for fname in os.listdir(os.path.join(self.path2database, 'pub2ref')))
        for ifile in tqdm(range(Nreffiles), desc='Removing Self-citations', disable= not show_progress):

            fname = os.path.join(self.path2database, 'pub2ref', "pub2ref{}.".format(ifile)+self.database_extension)
            refdf = self.read_file(fname)

            # get citing cited pairs with no common authors
            noselfcite = np.array([len(pub2authors[citingpid] & pub2authors[citedpid]) == 0 for citingpid, citedpid in refdf.values])

            # keep only citing-cited pairs without a common author
            refdf = refdf.loc[noselfcite]

            if preprocess:
                fname = os.path.join(self.path2database, fileprefix, fileprefix + '{}.'.format(ifile) + self.database_extension)
                self.save_data_file(refdf, fname, key = 'pub2ref')
            else:
                fullrefdf.append(refdf)

        if not preprocess:
            return pd.concat(fullrefdf)


    def compute_impact(self, preprocess=True, citation_horizons = [5,10], noselfcite = True):
        """
        Calculate several of the common citation indices.
            * 'Ctotal' : The total number of citations.
            * 'Ck' : The total number of citations within the first k years of publcation, for each k value specified by `citation_horizons`.
            * 'Ctotal_noself' : The total number of citations with self-citations removed.
            * 'Ck' : The total number of citations within the first k years of publcation with self-citations removed, for each k value specified by `citation_horizons`.

        Parameters
        ----------
        preprocess : bool, default True, Optional
            If True then the impact measures are saved in preprocessed files.

        citation_horizons : list, default [5,10], Optional
            The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

        noselfcite : Bool, default 'True', Optional
            If True then the noselfcitation pub2ref files are also processed.

        Returns
        -------
        DataFrame
            The impact DataFrame with at least two columns: 'PublicationId', 'Year', + citation columns
        

        |

        """

        # first load the publication year information
        pub2year = self.load_pub2year()

        # now get the reference list and merge with year info
        pub2ref = self.pub2ref

        # drop all citations that happend before the publication year
        pub2ref = pub2ref.loc[[pub2year.get(citingpid, 0) >= pub2year.get(citedpid, 0) for citingpid, citedpid in pub2ref[['CitingPublicationId', 'CitedPublicationId']].values]]

        # calcuate the total citations
        citation = groupby_count(pub2ref, colgroupby='CitedPublicationId', colcountby='CitingPublicationId', count_unique=True )
        citation.rename(columns={'CitingPublicationIdCount':'Ctotal', 'CitedPublicationId':'PublicationId'}, inplace=True)

        # go from the larest k down
        for k in np.sort(citation_horizons)[::-1]:

            # drop all citations that happend after the k
            #pub2ref = pub2ref.loc[pub2ref['CitingPublicationYear'] <= pub2ref['CitedPublicationYear'] + k]
            pub2ref = pub2ref.loc[[pub2year.get(citingpid, 0) <= pub2year.get(citedpid, 0) + k for citingpid, citedpid in pub2ref[['CitingPublicationId', 'CitedPublicationId']].values]]

            # recalculate the impact
            k_citation = groupby_count(pub2ref, colgroupby='CitedPublicationId', colcountby='CitingPublicationId', count_unique=True )
            k_citation.rename(columns={'CitingPublicationIdCount':'C{}'.format(k), 'CitedPublicationId':'PublicationId'}, inplace=True)

            citation = citation.merge(k_citation, how='left', on='PublicationId')

        # get the Cited Year
        citation['Year'] = [pub2year.get(pid, 0) for pid in citation['PublicationId'].values]


        if noselfcite:
            del pub2ref
            pub2ref = self.pub2refnoself

            # drop all citations that happend before the publication year
            pub2ref = pub2ref.loc[[pub2year.get(citingpid, 0) >= pub2year.get(citedpid, 0) for citingpid, citedpid in pub2ref[['CitingPublicationId', 'CitedPublicationId']].values]]

            # calcuate the total citations
            citation_noself = groupby_count(pub2ref, colgroupby='CitedPublicationId', colcountby='CitingPublicationId', count_unique=True )
            citation_noself.rename(columns={'CitingPublicationIdCount':'Ctotal_noself', 'CitedPublicationId':'PublicationId'}, inplace=True)

            # go from the larest k down
            for k in np.sort(citation_horizons)[::-1]:

                # drop all citations that happend after the k
                #pub2ref = pub2ref.loc[pub2ref['CitingPublicationYear'] <= pub2ref['CitedPublicationYear'] + k]
                pub2ref = pub2ref.loc[[pub2year.get(citingpid, 0) <= pub2year.get(citedpid, 0) + k for citingpid, citedpid in pub2ref[['CitingPublicationId', 'CitedPublicationId']].values]]

                # recalculate the impact
                k_citation = groupby_count(pub2ref, colgroupby='CitedPublicationId', colcountby='CitingPublicationId', count_unique=True )
                k_citation.rename(columns={'CitingPublicationIdCount':'C{}_noself'.format(k), 'CitedPublicationId':'PublicationId'}, inplace=True)

                citation_noself = citation_noself.merge(k_citation, how='left', on='PublicationId')

            citation = citation.merge(citation_noself, how='left', on='PublicationId')

        # set all nan to 0
        citation.fillna(0, inplace=True)

        if preprocess:

            if not os.path.exists(os.path.join(self.path2database, 'impact')):
                os.mkdir(os.path.join(self.path2database, 'impact'))

            for y, cdf in citation.groupby('Year', sort=True):
                fname = os.path.join(self.path2database, 'impact', 'impact{}.'.format(y) + self.database_extension)
                self.save_data_file(cdf, fname, key ='impact')

        else:
            return citation




    def precompute_teamsize(self, save2pubdf = True, show_progress=False):
        """
        Calculate the teamsize of publications, defined as the total number of Authors on the publication.

        Parameters
        ----------
        save2pubdf : bool, default True, Optional
            If True the results are appended to the preprocessed publication DataFrames.

        show_progress: bool, default False
            If True, display a progress bar for the count.

        Returns
        -------
        DataFrame
            TeamSize DataFrame with 2 columns: 'PublicationId', 'TeamSize'
        
        
        |

        """
        if show_progress or self.show_progress: 
            print("Starting TeamSize computation. \nLoading Data.")

        pub2authorseq = self.load_publicationauthoraffiliation(columns = ['PublicationId', 'AuthorId', 'AuthorSequence'],
            duplicate_subset = ['PublicationId', 'AuthorId'], dropna = ['PublicationId', 'AuthorId'])

        # register our pandas apply with tqdm for a progress bar
        tqdm.pandas(desc='TeamSize', disable= not show_progress)

        pub2teamsize = pub2authorseq.groupby('PublicationId', sort=False)['AuthorSequence'].progress_apply(lambda x: x.max()).astype(int).to_frame().reset_index().rename(columns={'AuthorSequence':'TeamSize'})

        if save2pubdf:
            if show_progress: print("Saving Teamsize.")
            append_to_preprocessed(pub2teamsize, self.path2database, 'publication')

        return pub2teamsize


    def precompute_yearly_citations(self, preprocess = True, show_progress=False):

        if show_progress:
            print("Starting Computation of Yearly Citations")

        # first load the publication year information
        pub2year = self.pub2year

        # now get the reference list and merge with year info
        pub2ref = self.pub2ref

        pub2ref['CitingYear'] = [pub2year.get(citingpid, 0) for citingpid in pub2ref['CitingPublicationId'].values]

        # drop all citations that happend before the publication year
        pub2ref = pub2ref.loc[[citingyear >= pub2year.get(citedpid, 0) for citingyear, citedpid in pub2ref[['CitingYear', 'CitedPublicationId']].values]]

        if show_progress:
            print("Yearly Citation Data Prepared")

        # calcuate the total citations
        citation = groupby_count(pub2ref, colgroupby=['CitedPublicationId', 'CitingYear'], colcountby='CitingPublicationId', count_unique=True )
        citation.rename(columns={'CitingPublicationIdCount':'YearlyCitations', 'CitedPublicationId':'PublicationId'}, inplace=True)

        # get the Cited Year
        citation['CitedYear'] = [pub2year.get(pid, 0) for pid in citation['PublicationId'].values]

        citation.sort_values(by=['CitedYear', 'CitedPublicationId', 'CitingYear'], inplace=True)

        if show_progress:
            print("Yearly Citations Found")

        if preprocess:
            if not os.path.exists(os.path.join(self.path2database, 'temporalimpact')):
                os.mkdir(os.path.join(self.path2database, 'temporalimpact'))

            for y, cdf in citation.groupby('CitedYear', sort=True):
                fname = os.path.join(self.path2database, 'temporalimpact', 'temporalimpact{}.'.format(y) + self.database_extension)
                self.save_data_file(cdf, fname, key ='temporalimpact')

            if show_progress:
                print("Yearly Citations Saved")

        else:
            return citation




# -*- coding: utf-8 -*-
"""
.. module:: distruption index
    :synopsis: Set of functions for finding the disruption index.

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """
import sys
import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

### Disruption
def disruption_index(pub2ref, focus_pub_ids = None, cite_window = None, ref_window = None, show_progress=False):
    """
    Calculate the disruption index as first proposed in :cite:`Funk2017disrupt` and used in :cite:`Wu2019teamsdisrupt`.
    We also include the windowed disruption index used in :cite:`Park2023timedisrupt`.

    Parameters
    ----------

    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param focus_pub_ids : numpy array
        A subset of publication ids to focus on for the disruption index.

    :param cite_window : list of two ints, default None
        If None, no citation window is applied.
        If [l, u] where, l,u are ints, then only citations whose year difference is greater than or equal to the lower bound l and 
            less than or equal to the upper bound u are used.  e.g. [0,5] uses citations within 5 years of publication (and not before).

    :param ref_window : list of two ints, default None
        If None, no reference window is applied.
        If [l, u] where, l,u are ints, then only references whose year difference is greater than or equal to the lower bound l and 
            less than or equal to the upper bound u are used.  e.g. [0,5] uses references within 5 years of publication (and not after).

    show_progress : bool, default False
        Show calculation progress. 

    Returns
    -------
    disruption : DataFrame
        A DataFrame with the disruption index for all (cited) publications or publications from the focus_pub_ids list.


    """
    if show_progress:
        print("Starting computation of disruption index.")

    if ref_window is None:
        reference_groups = pub2ref.groupby('CitingPublicationId', sort = False)['CitedPublicationId']
    else:
        ref_sub = [ ((y1-y2) >= ref_window[0] and (y1-y2) <=ref_window[1]) for y1,y2 in pub2ref[['CitingYear', 'CitedYear']].values]
        reference_groups = pub2ref.loc[ref_sub].groupby('CitingPublicationId', sort = False)['CitedPublicationId']

    if cite_window is None:
        citation_groups = pub2ref.groupby('CitedPublicationId', sort = False)['CitingPublicationId']
    else:
        cite_sub = [ ((y1-y2) >= cite_window[0] and (y1-y2) <=cite_window[1]) for y1,y2 in pub2ref[['CitingYear', 'CitedYear']].values]
        citation_groups = pub2ref.loc[cite_sub].groupby('CitingPublicationId', sort = False)['CitedPublicationId']

    if focus_pub_ids is None:
        if cite_window is None:
            focus_pub_ids = pub2ref['CitedPublicationId'].unique()
        else:
            focus_pub_ids = pub2ref.loc[cite_sub]['CitedPublicationId'].unique()

    def get_citation_groups(pid):
        try:
            return citation_groups.get_group(pid).values
        except KeyError:
            return np.array([])

    def _disruption_index(focusid):

        # if the focus publication has no references or citations, then it has a disruption of None
        try:
            focusref = reference_groups.get_group(focusid)
        except KeyError:
            return None

        try:
            citing_focus = citation_groups.get_group(focusid)
        except KeyError:
            return None


        # implementation 1: keep it numpy
        #cite2ref = reduce(np.union1d, [get_citation_groups(refid) for refid in focusref])
        #nj = np.intersect1d(cite2ref, citing_focus.values).shape[0]
        #nk = cite2ref.shape[0] - nj

        # implementation 2: but dicts are faster...
        cite2ref = {citeid:1 for refid in focusref for citeid in get_citation_groups(refid)}
        nj = sum(cite2ref.get(pid, 0) for pid in citing_focus.values )
        nk = len(cite2ref) - nj

        ni = citing_focus.shape[0] - nj

        return float(ni - nj)/(ni + nj + nk)

    disrupt = [[focusciting, _disruption_index(focusciting)] for focusciting
        in tqdm(focus_pub_ids, leave=True, desc='Disruption Index', disable= not show_progress) if get_citation_groups(focusciting).shape[0] > 0]

    return pd.DataFrame(disrupt, columns = ['PublicationId', 'DisruptionIndex'])




# -*- coding: utf-8 -*-
"""
.. module:: authormetrics
    :synopsis: Set of functions for the bibliometric analysis of authors

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import sys

import pandas as pd
import numpy as np

# determine if we are loading from a jupyter notebook (to make pretty progress bars)
if 'ipykernel' in sys.modules:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

from pyscisci.utils import zip2dict, check4columns, groupby_count, groupby_range, empty_mode
from pyscisci.methods.hindex import compute_hindex, compute_gindex
from pyscisci.methods.qfactor import compute_qfactor
from pyscisci.methods.productivitytrajectory import yearly_productivity_traj
from pyscisci.methods.diffusionscientificcredit import diffusion_of_scientific_credit
from pyscisci.methods.careertopics import career_cociting_network_topics
from pyscisci.methods.hotstreak import career_hotstreak

def author_productivity(pub2author=None, colgroupby = 'AuthorId', colcountby = 'PublicationId', show_progress=False):
    """
    Calculate the total number of publications for each author.

    Parameters
    ----------
    pub2author : DataFrame, default None, Optional
        A DataFrame with the author2publication information.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'PublicationId', Optional
        The DataFrame column with Publication Ids.  If None then the database 'PublicationId' is used.


    Returns
    -------
    DataFrame
        Productivity DataFrame with 2 columns: 'AuthorId', 'Productivity'

    """

    # we can use show_progress to pass a label for the progress bar
    if show_progress:
        show_progress='Author Productivity'

    if pub2author is None:
        pub2author = self.pub2author

    newname_dict = zip2dict([str(colcountby)+'Count', '0'], ['Productivity']*2)
    return groupby_count(pub2author, colgroupby, colcountby, count_unique=True, show_progress=show_progress).rename(columns=newname_dict)

def author_yearly_productivity(pub2author=None, colgroupby = 'AuthorId', datecol = 'Year', colcountby = 'PublicationId', show_progress=False):
    """
    Calculate the number of publications for each author in each year.

    Parameters
    ----------
    pub2author : DataFrame, default None, Optional
        A DataFrame with the author2publication information.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    datecol : str, default 'Year', Optional
        The DataFrame column with Year information.  If None then the database 'Year' is used.

    colcountby : str, default 'PublicationId', Optional
        The DataFrame column with Publication Ids.  If None then the database 'PublicationId' is used.

    Returns
    -------
    DataFrame
        Productivity DataFrame with 3 columns: 'AuthorId', 'Year', 'YearlyProductivity'

    """

    # we can use show_progress to pass a label for the progress bar
    if show_progress:
        show_progress='Yearly Productivity'

    if pub2author is None:
        pub2author = self.pub2author

    if not 'Year' in list(pub2author):
        pub2year = self.pub2year
        pub2author['Year'] = [pub2year.get(pid, None) for pid in pub2author['PublicationId'].values]
        pub2author.dropna(subset=['Year'], inplace=True)

    newname_dict = zip2dict([str(colcountby)+'Count', '0'], ['YearlyProductivity']*2)
    return groupby_count(pub2author, [colgroupby, datecol], colcountby, count_unique=True, show_progress=show_progress).rename(columns=newname_dict)

def author_career_length(pub2author = None, colgroupby = 'AuthorId', datecol = 'Year', show_progress=False):
    """
    Calculate the career length for each author.  The career length is the length of time from the first
    publication to the last publication.

    Parameters
    ----------
    pub2author : DataFrame, default None, Optional
        A DataFrame with the author2publication information.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    datecol : str, default 'Year', Optional
        The DataFrame column with Date information.  If None then the database 'Year' is used.

    Returns
    -------
    DataFrame
        Productivity DataFrame with 2 columns: 'AuthorId', 'CareerLength'

    """

    # we can use show_progress to pass a label for the progress bar
    if show_progress:
        show_progress='Career Length'

    newname_dict = zip2dict([str(datecol)+'Range', '0'], ['CareerLength']*2)
    return groupby_range(pub2author, colgroupby, datecol, show_progress=show_progress).rename(columns=newname_dict)

def author_startyear(pub2author = None, colgroupby = 'AuthorId', datecol = 'Year', show_progress=False):
    """
    Calculate the year of first publication for each author.

    Parameters
    ----------
    pub2author : DataFrame, default None, Optional
        A DataFrame with the author2publication information.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    datecol : str, default 'Year', Optional
        The DataFrame column with Date information.  If None then the database 'Year' is used.

    Returns
    -------
    DataFrame
        Productivity DataFrame with 2 columns: 'AuthorId', 'CareerLength'

    """

    newname_dict = zip2dict([str(datecol), '0'], ['StartYear']*2)
    return pub2author.groupby(colgroupby)[datecol].min().to_frame().reset_index().rename(columns=newname_dict)

def author_endyear(pub2author = None, colgroupby = 'AuthorId', datecol = 'Year', show_progress=False):
    """
    Calculate the year of last publication for each author.

    Parameters
    ----------
    pub2author : DataFrame, default None, Optional
        A DataFrame with the author2publication information.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    datecol : str, default 'Year', Optional
        The DataFrame column with Date information.  If None then the database 'Year' is used.

    Returns
    -------
    DataFrame
        Productivity DataFrame with 2 columns: 'AuthorId', 'CareerLength'

    """

    newname_dict = zip2dict([str(datecol), '0'], ['EndYear']*2)
    return pub2author.groupby(colgroupby)[datecol].max().to_frame().reset_index().rename(columns=newname_dict)


def author_productivity_trajectory(pub2author, colgroupby = 'AuthorId', datecol = 'Year', colcountby = 'PublicationId', show_progress=False):
    """
    Calculate the author yearly productivity trajectory.  See :cite:p:`way2017misleading`

    The algorithmic implementation can be found in :py:func:`metrics.compute_yearly_productivity_traj`.

    Parameters
    ----------
    pub2author : DataFrame, default None
        A DataFrame with the author2publication information.

    colgroupby : str, default 'AuthorId'
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    datecol : str, default 'Year'
        The DataFrame column with Date information.  If None then the database 'Year' is used.

    colcountby : str, default 'PublicationId'
        The DataFrame column with Publication Ids.  If None then the database 'PublicationId' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 5 columns: 'AuthorId', 't_break', 'b', 'm1', 'm2'

    """
    if not 'YearlyProductivity' in list(pub2author):
        yearlyprod = author_yearly_productivity(pub2author, colgroupby=colgroupby, datecol=datecol, colcountby=colcountby)
    else:
        yearlyprod = pub2author
    return yearly_productivity_traj(yearlyprod, colgroupby = colgroupby)

def author_hindex(pub2author, impact=None, colgroupby = 'AuthorId', colcountby = 'Ctotal', show_progress=False):
    """
    Calculate the author H-index.  See :cite:`hirsch2005index` for the derivation.

    The algorithmic implementation can be found in :py:func:`metrics.hindex`.

    Parameters
    ----------
    df : DataFrame, default None, Optional
        A DataFrame with the author2publication information.  If None then the database 'author2pub' is used.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'AuthorId', 'Hindex'

    """

    #df =

    if show_progress: print("Computing H-index.")
    if impact is None:
        df = pub2author
    else:
        df = pub2author.merge(impact[[colgroupby, colcountby]], on='PublicationId', how='left')

    return compute_hindex(df, colgroupby = colgroupby, colcountby = colcountby, show_progress=show_progress)

def author_gindex(pub2author, impact=None, colgroupby = 'AuthorId', colcountby = 'Ctotal', show_progress=False):
    """
    Calculate the author g-index.  See :cite:`hirsch2005index` for the derivation.

    The algorithmic implementation can be found in :py:func:`metrics.hindex`.

    Parameters
    ----------
    df : DataFrame, default None, Optional
        A DataFrame with the author2publication information.  If None then the database 'author2pub' is used.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'AuthorId', 'Hindex'

    """

    #df =

    if show_progress: print("Computing G-index.")
    if impact is None:
        df = pub2author
    else:
        df = pub2author.merge(impact[[colgroupby, colcountby]], on='PublicationId', how='left')

    return compute_gindex(df, colgroupby = colgroupby, colcountby = colcountby, show_progress=show_progress)

def author_hindex(pub2author, impact=None, colgroupby = 'AuthorId', colcountby = 'Ctotal', show_progress=False):
    """
    Calculate the author yearly productivity trajectory.  See :cite:`hirsch2005index` for the derivation.

    The algorithmic implementation can be found in :py:func:`metrics.hindex`.

    Parameters
    ----------
    df : DataFrame, default None, Optional
        A DataFrame with the author2publication information.  If None then the database 'author2pub' is used.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'AuthorId', 'Hindex'

    """

    #df =

    if show_progress: print("Computing H-index.")
    if impact is None:
        df = pub2author
    else:
        df = pub2author.merge(impact[[colgroupby, colcountby]], on='PublicationId', how='left')

    return compute_hindex(df, colgroupby = colgroupby, colcountby = colcountby, show_progress=show_progress)

def author_qfactor(pub2author, impact=None, colgroupby = 'AuthorId', colcountby = 'Ctotal', show_progress=False):
    """
    Calculate the author yearly productivity trajectory.  See :cite:`Sinatra2016qfactor` for the derivation.

    The algorithmic implementation can be found in :py:func:`metrics.qfactor`.

    Parameters
    ----------
    df : DataFrame, default None, Optional
        A DataFrame with the author2publication information.  If None then the database 'author2pub' is used.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'AuthorId', 'Hindex'

    """

    #df =

    if show_progress: print("Computing Q-factor.")
    if impact is None:
        df = pub2author
    else:
        df = pub2author.merge(impact[[colgroupby, colcountby]], on='PublicationId', how='left')

    return compute_qfactor(df, colgroupby = colgroupby, colcountby = colcountby, show_progress=show_progress)

def author_cindex(pub2author, impact=None, colgroupby = 'AuthorId', colcountby = 'Ctotal', show_progress=False):
    """
    Calculate the author c-index.  See :cite:`Waltman2008index` for the derivation.

    The number of citations for an author's most cited work.

    Parameters
    ----------
    df : DataFrame, default None, Optional
        A DataFrame with the author2publication information.  If None then the database 'author2pub' is used.

    colgroupby : str, default 'AuthorId', Optional
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'Ctotal', Optional
        The DataFrame column with Citation counts for each publication.  If None then the database 'Ctotal' is used.

    Returns
    -------
    DataFrame
        Trajectory DataFrame with 2 columns: 'AuthorId', 'Hindex'

    """

    #df =

    if show_progress: print("Computing c-index.")
    if impact is None:
        df = pub2author
    else:
        df = pub2author.merge(impact[[colgroupby, colcountby]], on='PublicationId', how='left')

    return compute_qfactor(df, colgroupby = colgroupby, colcountby = colcountby, show_progress=show_progress)

def author_top_field(pub2author, colgroupby = 'AuthorId', colcountby = 'FieldId', 
    fractional_field_counts = False, show_progress=False):
    """
    Calculate the most frequent field in the authors career.

    Parameters
    ----------
    pub2author : DataFrame
        A DataFrame with the author2publication field information.

    colgroupby : str, default 'AuthorId'
        The DataFrame column with Author Ids.  If None then the database 'AuthorId' is used.

    colcountby : str, default 'FieldId'
        The DataFrame column with Citation counts for each publication.  If None then the database 'FieldId' is used.

    fractional_field_counts : bool, default False
        How to count publications that are assigned to multiple fields:
            - If False, each publication-field assignment is counted once.
            - If True, each publication is counted once, contributing 1/#fields to each field.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: 'AuthorId', 'TopFieldId'

    """

    check4columns(pub2author, [colgroupby, 'PublicationId', colcountby])

    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Author Top Field', disable= not show_progress)

    if not fractional_field_counts:
        author2field = pub2author.groupby(colgroupby)[colcountby].progress_apply(empty_mode)
    else:
        # first calculate how many fields each publication maps too
        pub2nfields = groupby_count(pub2author, colgroupby='PublicationId', colcountby=colcountby)

        # each pub2field mapping is weighted by the number of fields for the publication
        pub2nfields['PublicationWeight'] = 1.0/pub2nfields['PublicationIdCount']
        del pub2nfields[str(colcountby)+'Count']

        # merge counts
        author2field = pub2author.merge(pub2nfields, how='left', on='PublicationId')

        # custom weighted mode based on 
        def weighted_mode(adf):
            p = adf.groupby(colcountby)['PublicationWeight'].sum()
            return p.idxmax()

        # now take the weighted mode for each groupby column
        author2field = author2field.groupby(colgroupby).progress_apply(weighted_mode)

    newname_dict = zip2dict([str(colcountby), '0'], ['Top' + str(colcountby)]*2)
    return author2field.to_frame().reset_index().rename(columns=newname_dict)


def author_hotstreak(pub2author, colgroupby = 'AuthorId', citecol = 'c10',datecol='Year',  maxk=1, l1_lambda=1.0, show_progress=False):
    """
    Identify hot streaks in author careers :cite:`liu2018hotstreak'.

    TODO: this is an interger programming problem.  Reimplement using an interger solver.
    Right now just using a brut force search (very inefficient)!
    
    Parameters
    ----------
    pub2author : DataFrame
        The author publication history for all authors.

    colgroupby : str, default 'AuthorId'
        The column with Author information.

    citecol : str, default 'c10'
        The column with publication citation information.

    datecol : str, default 'Year'
        The column with publication date/year information.
    
    max_k : int, default 1
        The maximum number of hot streaks to search for in a career. Should be 1 or 2.
    
    l1_lambda : float, default 1.0
        The l1 regularization for the number of streaks.  
        Note, the authors never define the value they used for this in the SI.
        
    Returns
    ----------
    lsm_err : float
        The least square mean error of the model plus the l1-regularized term for the number of model coefficients.

    streak_loc : array
        The index locations for the hot streak start and end locations.
    """

    # register our pandas apply with tqdm for a progress bar
    tqdm.pandas(desc='Author HotStreak', disable= not show_progress)

    pub2author = pub2author.sort_values(by=[colgroupby, datecol]).reset_index(drop=True)

    hotstreak = pub2author.groupby(colgroupby).progress_apply(lambda x: career_hotstreak(x, citecol=citecol, maxk=maxk, l1_lambda = l1_lambda))
    return hotstreak.reset_index().rename(columns={'level_1':'StreakNumber'})

class pySciSciMetricError(Exception):
    """
    Base Class for metric errors.
    """
    def __str__(self, msg=None):
        if msg is None:
            return 'pySciSci metric error.'
        else:
            return msg



# -*- coding: utf-8 -*-
"""
.. module:: interdisciplinary
    :synopsis: Set of functions for typical interdisciplinary analysis

.. moduleauthor:: Alex Gates <ajgates42@gmail.com>
 """

import pandas as pd
import numpy as np

from ..utils import isin_sorted, zip2dict, check4columns, simpson, simpson_finite, shannon_entropy



def simpson_interdisciplinarity(pub2ref, pub2field, focus_pub_ids=None,
    citation_direction='references', finite_correction=False, show_progress=False):
    """
    Calculate the Simpson index as a measure of a publication's interdisciplinarity.
    See :cite:`stirling20` for the definition.

    Parameters
    ----------
    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2field : DataFrame
        A DataFrame with the field information for each Publication.

    :param focus_pub_ids : numpy array or list, default None
        A list of the PublicationIds to calculate interdisciplinarity.

    :param finite_correction : bool, default False
        Whether to apply the correction for a finite sample.

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: 'PublicationId', 'Simpsons'

    """

    # now we map citing and cited to the source and target depending on which diretion was specified by `citation_direction'
    if citation_direction == 'references':
        pub2ref_rename_dict = {'CitedPublicationId':'TargetId', 'CitingPublicationId':'SourceId'}
        year_col = 'CitingYear'
    elif citation_direction == 'citations':
        pub2ref_rename_dict = {'CitedPublicationId':'SourceId', 'CitingPublicationId':'TargetId'}
        year_col = 'CitedYear'

    required_columns = ['CitedPublicationId', 'CitingPublicationId']
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].rename(columns=pub2ref_rename_dict)

    check4columns(pub2field, ['PublicationId', 'FieldId'])

    # merge the references to the fields for the target fields
    pub2ref = pub2ref.merge(pub2field, how='left', left_on='TargetId', 
        right_on='PublicationId').rename(columns={'FieldId':'TargetFieldId'})
    del pub2ref['PublicationId']

    pub2ref = pub2ref.dropna()
    
    if finite_correction:
        simpdf = 1-pub2ref.groupby('SourceId')['TargetFieldId'].apply(simpson_finite)
    else:
        simpdf = 1-pub2ref.groupby('SourceId')['TargetFieldId'].apply(simpson)

    simpdf = simpdf.to_frame().reset_index().rename(
        columns={'TargetFieldId':'SimpsonInterdisciplinarity', 'SourceId':'PublicationId'})
    
    return simpdf


def shannon_interdisciplinarity(pub2ref, pub2field, focus_pub_ids=None,
    citation_direction='references', normalized=False, K=None, show_progress=False):
    """
    Calculate the Shannon entropy as a measure of a publication's interdisciplinarity.
    See :cite:`stirling20` for the definition.

    Parameters
    ----------
    :param pub2ref : DataFrame
        A DataFrame with the citation information for each Publication.

    :param pub2field : DataFrame
        A DataFrame with the field information for each Publication.

    :param focus_pub_ids : numpy array or list, default None
        A list of the PublicationIds to calculate interdisciplinarity.

    :param temporal : bool, default False
        If True, compute the distance matrix using only publications for each year.

    :param normalized : bool, default False
        If True, use the normalized entorpy bounded by the number of observed fields
        or K if not None.
    
    :param K : int, default None
        The maximum number of fields to consider.

    :param show_progress : bool, default False
        If True, show a progress bar tracking the calculation.

    Returns
    -------
    DataFrame
        DataFrame with 2 columns: 'PublicationId', 'Simpsons'

    """

    # now we map citing and cited to the source and target depending on which diretion was specified by `citation_direction'
    if citation_direction == 'references':
        pub2ref_rename_dict = {'CitedPublicationId':'TargetId', 'CitingPublicationId':'SourceId'}
        year_col = 'CitingYear'
    elif citation_direction == 'citations':
        pub2ref_rename_dict = {'CitedPublicationId':'SourceId', 'CitingPublicationId':'TargetId'}
        year_col = 'CitedYear'

    required_columns = ['CitedPublicationId', 'CitingPublicationId']
    check4columns(pub2ref, required_columns)
    pub2ref = pub2ref[required_columns].rename(columns=pub2ref_rename_dict)

    check4columns(pub2field, ['PublicationId', 'FieldId'])

    if K is None:
        K = pub2field['FieldId'].nunique()

    # merge the references to the fields for the target fields
    pub2ref = pub2ref.merge(pub2field, how='left', left_on='TargetId', 
        right_on='PublicationId').rename(columns={'FieldId':'TargetFieldId'})
    del pub2ref['PublicationId']

    pub2ref = pub2ref.dropna()
    
    shan_inter = pub2ref.groupby('SourceId')['TargetFieldId'].apply(shannon_entropy)
    shan_inter = shan_inter.to_frame().reset_index().rename(
        columns={'TargetFieldId':'ShannonInterdisciplinarity', 'SourceId':'PublicationId'})

    if normalized:
        shan_inter['ShannonInterdisciplinarity'] = shan_inter['ShannonInterdisciplinarity']/np.log(K)
    
    return shan_inter

